#pragma once

#include <string>
#include <vector>
#include <map>
#include <database/SqlQuery.h>
#include <database/SqlPreparedQuery.h>
#include <database/export.h>

namespace pdp {

/**
 * The pdp SQL Database class
 *
 * @ingroup database
 */
class PDPDATABASE_EXPORT SqlDatabase
{
public:
  /**
   * Enumerated type defining the supported sql database technologies
   */
  enum SqlDatabaseImplementation
  {
    SqliteImplementation = 0,
    PostgresImplementation
  };

  /**
   * SqlDatabase object destructor.
   */
  virtual ~SqlDatabase() {}

  /**
   * Returns the database implementation
   */
  virtual SqlDatabase::SqlDatabaseImplementation implementation() const = 0;
  
  /**
   * Provides the data type string used for representing blob types in the given
   * database type
   */
  virtual std::string blobTypeString() const = 0;

  /**
   * Provides the data type string used for representing Uuid types.  
   */
  virtual std::string uuidTypeString() const = 0;

  /**
   * Returns the datatype to use for Json serialization in the given database.
   */
  virtual std::string jsonTypeString() const = 0;

  /**
   * Checks the schema to see if the table exists.  If the database is not
   * connected, this function returns a false.
   */
  virtual bool tableExists(
    const std::string& tableName) = 0;

  /**
   * Checks the schema to see if the requested column exists in the table. If
   * the database is not connected, this function returns false.
   */
  virtual bool tableColumnExists(
    const std::string& tableName,
    const std::string& columnName) = 0;

  /**
   * Returns a complete list of column names
   */
  virtual std::vector<std::string> columnNames(
    const std::string& tableName) = 0;

  /**
   * Allows a database to specify if it supports the Sql command "SET
   * TRANSACTION ISOLATION LEVEL".  Most database support this statement, and
   * will all allow you to upgrade the isolation level at the beginning of a
   * transaction.  Some will allow the use of this statement to upgrade the
   * transaction isolation level for the remaining database connection or the
   * database as a whole.  However that is where the database support for this
   * statement diverge
   * */
  virtual bool supportsTransactionIsolationLevel() const = 0;

  /**
   * Open an existing SqlDatabase. The @a connectionInfo string contains the command
   * string and is specific to the implementation.
   */
  virtual bool open(
    const std::string& connectionInfo,
    const std::string& password,
    bool enableNotification,
    bool applyLocalEncryption,
    bool inMemory) = 0;

  /**
   * Checks that the connection is currently valid and attempts to re-establish it
   * if it is not.  If <code>false</code> is returned, then the error message 
   * should reflect the implementation-specific reason.
   */
  virtual bool ensureConnected() = 0;

  /**
   * Returns @a true if the database is currently open.
   */
  virtual bool isOpen() const = 0;

  /**
   * Close the SqlDatabase.
   */
  virtual void close() = 0;

  /**
   * Re-assigns the encryption key to an existing database so that it can be
   * opened with the @a open command.
   */
  virtual bool rekey(
    const std::string& connectionInfo,
    const std::string& password) = 0;

  /**
   * Returns the error message from the last executed method.
   */
  virtual const std::string errorMessage() const = 0;

  virtual SqlQuery* createQuery() = 0;

  /**
   * Creates a prepared statement and returns a pointer to the object.  Note
   * ownership is retained by the SqlDatabase and this pointer should not be 
   * freed. The returned object will be valid until the database connection
   * has been terminated, at which point it will be deleted.  Sequential 
   * calls to the function with the same query will return the same object, 
   * since the prepared queries are cached by the SqlDatabase.
   */
  SqlPreparedQuery* createPreparedQuery(
    const std::string& sql);

  /**
   * This function deletes an SQL database.
   */
  virtual bool deleteDatabase(
    const std::string& databaseUsername,
    const std::string& databasePassword,
    const std::string& databaseName,
    const std::string& databasePath,
    const std::string& commandsIp,
    int commandsPort,
    std::string& errorMessage) = 0;

  /**
   * This function renames an SQL database.
   */
  virtual bool renameDatabase(
    const std::string& databaseUsername,
    const std::string& databasePassword,
    const std::string& databaseName,
    const std::string& databasePath,
    const std::string& commandsIp,
    int commandsPort,
    const std::string& newDatabaseName,
    std::string& errorMessage) = 0;

  /**
   * This function returns true if the database supports transactions.
   */
  virtual bool supportsTransactions() { return true; }

protected: 
  explicit SqlDatabase() {}

  void clearPreparedQueries();

  /**
   * Template method for derived classes to implement.  This method will be used
   * to cause the derived class to create a Prepared query for the sql
   * requested if it hasn't already been generated.  The QObject passed in must
   * be the owner of the generated object, so that its destruction will delete
   * the SqlPreparedQuery provided.
   */
  virtual SqlPreparedQuery* doCreatePreparedQuery(
    const std::string& sql) = 0;

private:
  mutable std::map<std::string, SqlPreparedQuery*> m_prepQueries;
};

} // namespace pdp

