#pragma once

#include <string>
#include <vector>
#include <utility>
#include <boost/variant.hpp>
#include <utilities/stringUtils.h>
#include <database/SqlResult.h>
#include <database/export.h>

namespace pdp {

/**
 * Prepared Query class.  This class sets up a template pattern
 * that will allow us to write a PreparedQuery class for any database.  
 * Data is stored in the base class and accessed through non-virtual functions
 * for optimized access.  Every derived class has a prepare() method to force
 * preparation of the query, however after constructing the query, running a
 * bindAndExecute will prepare the query first if necessary.  
 *
 * This class expects the SQL string to have named parameters in the form of
 * ':namedParameter'.  These parameters are then bound to values passed in at
 * execution time.  The class ensures that the caller provides all the binds
 * necessary by the SQL provided, and will fail is these requirements are not
 * meet. 
 *
 * Note that this class outputs to the console when the error is caused by an
 * improper use of the API.  This includes things such as mismatch in the named
 * parameters provided in the sql vs the binds.  
 *
 * The class expects its lifetime to be contolled by the owner class.  Calling
 * this classes functions after a database disconnect will
 * likely throw an exception, so the lifetime of the queries should match that
 * of the database.  The user should not delete this class manually as that will
 * cause an exception.
 *
 * @ingroup database
 */
class PDPDATABASE_EXPORT SqlPreparedQuery
{
public:
  typedef boost::variant<
    void*, /**< used to identify query with no variables */
    std::vector<unsigned char>, /**< Blob */
    int, /**< Integer */
    double, /**< Double percision float */
    float, /**< Single percision float */
    std::chrono::system_clock::time_point, /**< Date / Time data element */
    std::string, /**< String */
    bool> DatabaseDataTypes;

  typedef std::pair<std::string, DatabaseDataTypes> BindType;

  typedef std::vector<BindType> BindList;

  typedef pdp::SqlResult::QueryRow QueryRow;

  /**
   * SqlPreparedQuery object destructor.
   */
  virtual ~SqlPreparedQuery();

  /**
   * Attempts to create the prepared statement and returns false on a failure.
   * Accessing the errorMessage will show why the preparation failed.  Note this
   * function doesn't have to be called directly.  Instead when the first call
   * to bindAndExecute takes place, prepare will be called if necessary.  Note
   * if the statement is already prepared, this function simply returns true.
   *
   * This function extracts the list of named parameters from the SQL provided
   * at construction.
   *
   * This function provides a means to determine if there is a syntax error with
   * the sql apart from any issues with the bound data.
   */
  bool prepare();

  /**
   * Bind the required values and execute the query.  This function takes a
   * vector of pairs with the named placeholder ":ExampleName", with the value
   * for that field stored in a DatabaseDataTypes.  This allows binding of binary blob,
   * data types as well as integers, floats, doubles, and strings.  This
   * function will verify that each named placeholder in the SQL is provided in
   * the bind list.  Duplicate named parameters are not allowed.  
   *
   * This function will perform a prepare if the statement hasn't already been
   * prepared.  It will return false if the prepare fails, if there is a syntax
   * error in the binds, or if the execution of the query with the binds fails.
   * See the errorMessage() for details on what went wrong.
   *
   * For developers this function will also echo to the console usage errors
   * instead of throwing exceptions.
   */
  SqlResult bindAndExec(
    const SqlPreparedQuery::BindList& binds);

  /**
   * This is a convenience function when only one bind is necessary.
   *
   * It Binds the required value and execute the query.  This function takes a
   * input in the form of a named placeholder ":ExampleName", with the value
   * for that field stored in a DatabaseDataTypes.  This allows binding of binary blob,
   * data types as well as integers, floats, doubles, and strings.  This
   * function will verify that each named placeholder in the SQL is provided in
   * the bind list.   
   *
   * This function will perform a prepare if the statement hasn't already been
   * prepared.  It will return false if the prepare fails, if there is a syntax
   * error in the binds, or if the execution of the query with the binds fails.
   * See the errorMessage() for details on what went wrong.
   *
   * For developers this function will also echo to the console usage errors
   * instead of throwing exceptions.
   */
  SqlResult bindAndExec(
   const SqlPreparedQuery::BindType& bind);

  /**
   * Used to execute a query without binding any data to it.  Basically calls
   * bindAndExecute with an empty list.
   */
  SqlResult exec();
  
  /**
   * Returns the error message from the prepare.
   */
  const std::string errorMessage() const
  { return m_errorMessage; }

  const SqlResult& lastResult() const
  { return m_lastResult; }

protected:
  /**
   * This constructor is called by the derived classes with the SQL string of 
   * the query.  This template expects the string to include named place holders
   * that values will later be bound to at execution time.  The format of the
   * placeholders is ':placeHolder'.  The name must start with a ':' and may
   * contain any word character including '_'.  
   *
   * Note this class expects an QObject owner, and it is expected that the
   * parent is responsible for the lifetime of the prepared query.  The prepared
   * query objects should be deleted if the database is disconnected as calling
   * it after such a point will likely cause a crash.
   */
  explicit SqlPreparedQuery(
    const std::string& originalSQL);

  /**
   * Creates the prepared statement and returns false on a failure.
   * Accessing the errorMessage will show why the preparation failed.  All
   * derived class must define this method for the template pattern to work
   */
  virtual bool doPrepare() = 0;

  /**
   * This function performs database specific commands to perform the bind.  All
   * derived classes must define this method for the template pattern to work.
   */
  virtual bool doBindAndExecute(
    const SqlPreparedQuery::BindList& binds, 
    pdp::SqlResult& data) = 0;

  /**
   * Used by the derived class to set any error messages received.
   */
  void setErrorMessage(
    const std::string& message)
  { m_errorMessage = message; }

  /**
   * Provides the list of placeholders that were used for preparation.  Placeholders
   * not in the list will be ignored in the original SQL.
   */
  const std::vector<std::string>& placeholders()
  { return m_placeholders; }

  const std::string& originalSQL() const
  { return m_originalSql; }
  
  bool isInsert() const
  { return m_isInsert; }

private:
  bool m_prepared;
  std::string m_originalSql;
  std::vector<std::string> m_placeholders; //list of placeholders, order matters
  std::string m_errorMessage;
  bool m_isInsert;
  SqlResult m_lastResult;
};

PDPDATABASE_EXPORT void extractLabels(
  const std::vector<std::pair<std::string, SqlPreparedQuery::DatabaseDataTypes>>& binds,
  std::vector<std::string>& labels);

} // namespace pdp

