#pragma once

#include <vector>
#include <database/SqlValue.h> 
#include <database/export.h>

namespace pdp {

/**
 *
 * @ingroup database
 */
class PDPDATABASE_EXPORT SqlQuery
{
public:
  /**
   * Struct for holding a row query result
   */
  struct QueryRow
  { std::vector<SqlValue> values; };

  /**
   * SqlQuery object destructor.
   */
  virtual ~SqlQuery();

  /**
   * Executes in the input SQL command @a sqlCommand on the query's database. After
   * and exec call returns the results of any SQL SELECT commands will be contained
   * in the query
   */
  virtual bool exec(
    const std::string& sqlCommand) = 0;

  /**
   * Increment the results to point to the next results row
   */
  virtual bool next() const = 0;

  /**
   * Returns the value at index @a idx for the current result row
   */
  virtual const SqlValue& value(
    int idx) const = 0;

  /**
   * Returns the value at row @a row and column @a col for the current result
   */
  virtual const SqlValue& value(
  	unsigned int row,
    unsigned int col) const = 0;

  /**
   * This method returns the number of rows obtained from the query.
   */
  virtual unsigned int numRows() const = 0;

  /**
   * Provides the number of columns retrieved in the query.
   */
  virtual unsigned int numColumns() const = 0;

  /**
   * Number of rows affected by the last SQL operation.  This can be used
   * to determine is a UPDATE actually matched any rows to update or to count
   * the number of rows added by an insert.
   */
  virtual unsigned int rowsAffected() const = 0;

  /**
   * Clears the query of any results it currently contains
   */
  virtual void clear() = 0;

  /**
   * Returns the error message from the last executed method.
   */
  virtual const std::string errorMessage() const = 0;
};

} // namespace pdp

