/** @file
 *
 * Color classes and functions
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <utilities/stringUtils.h>
#include <elements/export.h>

namespace pdp {

/**
 * Represents a color using 8 bit unsigned red, green and blue components,
 * and an 8 bit unsigned alpha blending value.
 *
 * @ingroup elements
 */
class PDPELEMENTS_EXPORT Color
{
public:
  /**
   * Constructs a Color with a default value of black.
   */
  Color();

  /**
   * Constructs a Color from the input specification
   */
  explicit Color(
    unsigned char red,
    unsigned char green,
    unsigned char blue,
    unsigned char alpha = 255);

  /**
   * Returns the color red component
   */
  unsigned char red() const;

  /**
   * Sets the color red component
   */
  void setRed(
    unsigned char red);

  /**
   * Returns the color green component
   */
  unsigned char green() const;

  /**
   * Sets the color green component
   */
  void setGreen(
    unsigned char green);

  /**
   * Returns the color blue component
   */
  unsigned char blue() const;

  /**
   * Sets the color blue component
   */
  void setBlue(
    unsigned char blue);

  /**
   * Returns the color alpha component
   */
  unsigned char alpha() const;

  /**
   * Sets the color alpha component
   */
  void setAlpha(
    unsigned char alpha);

  /**
   * Returns the highlight for the color
   */
  const Color highlightColor() const;

  /**
   * Returns the complement for the color
   */
  const Color complement() const;

  /**
   * Assigns the color value by alpha blending the two input colors
   */
  inline void blend(
    const Color& color1,
    const Color& color2);

  /**
  * Gets the "raw" (unsigned int) internal value.
  */
  unsigned int rawValue() const
  {
    return m_value;
  }

  /**
  * Sets the "raw" (unsigned int) internal value.
  */
  void fromRawValue(
    unsigned int value)
  {
    m_value = value;
  }

  /**
   * Equal operator
   */
  bool operator == (
    const Color& color) const;

  /**
   * Not equal operator
   */
  bool operator != (
    const Color& color) const;

private:
  unsigned int m_value;
};

inline void Color::blend(
  const Color& color1,
  const Color& color2)
{
  unsigned char* dest = reinterpret_cast<unsigned char*>(&m_value);
  const unsigned char* src1 = reinterpret_cast<const unsigned char*>(&(color1.m_value));
  const unsigned char* src2 = reinterpret_cast<const unsigned char*>(&(color2.m_value));
  float alpha1 = (float)src1[3] / 255.0f;
  float alpha2 = (float)src2[3] / 255.0f;
  dest[0] = (unsigned char)((float)(src1[0]) * alpha1 +
                            (float)src2[0] * (1.0 - alpha1) * alpha2);
  dest[1] = (unsigned char)((float)(src1[1]) * alpha1 +
                            (float)src2[1] * (1.0 - alpha1) * alpha2);
  dest[2] = (unsigned char)((float)(src1[2]) * alpha1 +
                            (float)src2[2] * (1.0 - alpha1) * alpha2);
  dest[3] = (src1[3] > src2[3]) ? src1[3] : src2[3];
}

inline unsigned char rgbToGreyscale(
  const Color& color)
{
  unsigned char grey = (unsigned char)(0.299 * color.red() + 0.587 * color.green() + 0.114 * color.blue());
  return grey;
}

} // namespace pdp



