/** @file
 *
 * Line segments composed of vectors with arbitrary element type.
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <assert.h>
#include <ostream>
#include <elements/VectorR.h>
#include <elements/export.h>

namespace pdp
{
/**
 * The LineSegmentR template class represents line segments composed of
 * fixed-size, real vectors.  The dimension and element type of the line
 * segment are specified as part of the type, and are thus fixed at compile
 * time.
 *
 * @a N is the number of elements, which must be greater than zero.
 * @a E is the element type, typically @c double, @c float, or @c int.
 *
 * @ingroup elements
 */
template<unsigned int N, typename E>
class LineSegmentR
{
public:
  /**
   * The number of dimensions in line segments of this type.
   */
  static const unsigned int SIZE = N;

  /**
   * Constructs a line segment whose elements are uninitialized.
   */
  inline LineSegmentR() {}

  /**
   * Destroys a line segment.
   */
  inline ~LineSegmentR() {}

  /**
   * Constructs a line segment based on the @a start and @a end input vectors.
   */
  LineSegmentR(
    const VectorR<N, E>& start,
    const VectorR<N, E>& end);

  /**
   * Constructs a copy of @a l.
   */
  LineSegmentR(
    const LineSegmentR<N, E>& l);

  /**
   * Assigns this LineSegmentR from another LineSegmentR @a l.
   */
  LineSegmentR<N, E>& operator=(const LineSegmentR<N, E>& l);

  /**
   * Returns a constant reference to the start vector of the line segment.
   */
  const VectorR<N, E>& start() const
  { return m_start; }

  /**
   * Returns a constant reference to the end vector of the line segment.
   */
  const VectorR<N, E>& end() const
  { return m_end; }

  /**
   * Sets the start of the line segment to the input vector.
   */
  void setStart(
    const VectorR<N, E>& start)
  { m_start = start; }

  /**
   * Sets the end of the line segment to the input vector.
   */
  void setEnd(
    const VectorR<N, E>& end)
  { m_end = end; }

  /**
   * Sets the start and end of the line segment to the input vectors.
   */
  void set(
    const VectorR<N, E>& start,
    const VectorR<N, E>& end)
  {
    m_start = start;
    m_end = end;
  }

  /**
   * Returns true if this line segment equals input line segment @a l.
   */
  bool operator==(
    const LineSegmentR& l) const;

  /**
   * Returns true if this line segment is not equal to input line segment @a l.
   */
  bool operator!=(
    const LineSegmentR& l) const
  { return !(*this == l); }

  //---------------------------------------------------------------------------
  // Friend functions
  //---------------------------------------------------------------------------
  /**
   * The method gets the next LineSegmentR from the input stream and
   * decodes it from little endian format.
   */
  friend void getLineSegmentR(
    std::istream& inputStream,
    LineSegmentR<N, E>& l)
  {
    getVectorR(inputStream, l.m_start);
    getVectorR(inputStream, l.m_end);
  }

  /**
   * The method encodes the input LineSegmentR into little endian format
   * and writes it to the output stream.
   */
  friend void putLineSegmentR(
    const LineSegmentR<N, E>& l,
    std::ostream& outputStream)
  {
    putVectorR(l.m_start, outputStream);
    putVectorR(l.m_end, outputStream);
  }

private:
  // member variables
  VectorR<N, E> m_start;
  VectorR<N, E> m_end;
};

//---------------------------------------------------------------------------
// Related functions
//---------------------------------------------------------------------------
/**
 * Reports the value of the line segment @a l in text format to
 * the output stream @a out
 *
 * @relates LineSegmentR
 */
template<unsigned int N, typename E>
std::ostream& operator<<(
  std::ostream& out,
  const LineSegmentR<N, E> & l);

/**
 * Returns the length of the line segment.
 *
 * @relates LineSegmentR
 */
template<unsigned int N, typename E>
double length(
  const LineSegmentR<N, E>& l);

/**
 * Computes the distance from the input point @a p to the input
 * line segment @a l. The function computes both the distance to the
 * segment @a segmentDistance and the distance to the line @a lineDistance assuming
 * that the line has infinite extent.
 *
 * @relates LineSegmentR
 */
template<typename E>
void distanceFromPointToLineSegment(
	const VectorR<2, E>& p,
  const LineSegmentR<2, E>& l,
  double& segmentDistance,
  double& lineDistance);

template<unsigned int N, typename E>
void projectPointToLineSegment(
  const VectorR<N, E>& x0,
  const LineSegmentR<N, E>& l,
  VectorR<N, E>& pProjection,
  double& t);

/**
 * Returns @a true if the start and end vectors of the two line segments
 * are both within @a epsilon.
 *
 * @relates LineSegmentR
 */
template<unsigned int N, typename E>
bool almostEqual(
  const LineSegmentR<N, E>& l0,
  const LineSegmentR<N, E>& l1,
  double epsilon = PDP_EPSILON);

/**
 * Constructs and returns a Vector from the line segment start to the
 * line segment end.
 *
 * @relates LineSegmentR
 */
template<unsigned int N, typename E>
inline VectorR<N, E> makeSpanVector(
  const LineSegmentR<N, E>& l)
{
  VectorR<N, E> span;
  sub(l.end(), l.start(), span);
  return span;
}

// Short type names for common line segment specializations.
typedef LineSegmentR<2, int> LineSegmentR2i;
typedef LineSegmentR<2, unsigned int> LineSegmentR2ui;
typedef LineSegmentR<2, float> LineSegmentR2f;
typedef LineSegmentR<2, double> LineSegmentR2d;
typedef LineSegmentR<3, int> LineSegmentR3i;
typedef LineSegmentR<3, unsigned int> LineSegmentR3ui;
typedef LineSegmentR<3, float> LineSegmentR3f;
typedef LineSegmentR<3, double> LineSegmentR3d;
typedef LineSegmentR<4, int> LineSegmentR4i;
typedef LineSegmentR<4, unsigned int> LineSegmentR4ui;
typedef LineSegmentR<4, float> LineSegmentR4f;
typedef LineSegmentR<4, double> LineSegmentR4d;

} // namespace pdp

#include <elements/LineSegmentR.t>


