/** @file
 * 
 * Math utility classes and functions
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <iostream>
#include <elements/VectorR.h>

namespace pdp {

typedef float MatrixR4x4f[4][4];

static inline void matrixR4x4f_identity(
  MatrixR4x4f& M)
{
  int i, j;
  for(i=0; i<4; ++i)
    for(j=0; j<4; ++j)
      M[i][j] = i==j ? 1.f : 0.f;
}

static inline void matrixR4x4f_assign(
  MatrixR4x4f& M,
  float value00,
  float value01,
  float value02,
  float value03,
  float value10,
  float value11,
  float value12,
  float value13,
  float value20,
  float value21,
  float value22,
  float value23,
  float value30,
  float value31,
  float value32,
  float value33)
{
  M[0][0] = value00;
  M[1][0] = value01;
  M[2][0] = value02;
  M[3][0] = value03;
  M[0][1] = value10;
  M[1][1] = value11;
  M[2][1] = value12;
  M[3][1] = value13;
  M[0][2] = value20;
  M[1][2] = value21;
  M[2][2] = value22;
  M[3][2] = value23;
  M[0][3] = value30;
  M[1][3] = value31;
  M[2][3] = value32;
  M[3][3] = value33;
}

static inline void matrixR4x4f_dup(
  MatrixR4x4f& M,
  const MatrixR4x4f& N)
{
  int i, j;
  for(i=0; i<4; ++i)
    for(j=0; j<4; ++j)
      M[i][j] = N[i][j];
}

static inline void matrixR4x4f_row(
  VectorR4f& r,
  const MatrixR4x4f& M,
  int i)
{
  int k;
  for(k=0; k<4; ++k)
    r[k] = M[k][i];
}

static inline void matrixR4x4f_col(
  VectorR4f& r,
  const MatrixR4x4f& M,
  int i)
{
  int k;
  for(k=0; k<4; ++k)
    r[k] = M[i][k];
}

static inline void matrixR4x4f_transpose(
  MatrixR4x4f& M,
  const MatrixR4x4f& N)
{
  int i, j;
  for(j=0; j<4; ++j)
    for(i=0; i<4; ++i)
      M[i][j] = N[j][i];
}

static inline void matrixR4x4f_add(
  MatrixR4x4f& M,
  const MatrixR4x4f& a,
  const MatrixR4x4f& b)
{
  int i, j;
  for(j=0; j<4; ++j)
    for(i=0; i<4; ++i)
      M[j][i] = a[j][i] + b[j][i];
}

static inline void matrixR4x4f_sub(
  MatrixR4x4f& M,
  const MatrixR4x4f& a,
  const MatrixR4x4f& b)
{
  int i, j;
  for(j=0; j<4; ++j)
    for(i=0; i<4; ++i)
      M[j][i] = a[j][i] - b[j][i];
}

static inline void matrixR4x4f_scale(
  MatrixR4x4f& M,
  const MatrixR4x4f& a,
  float k)
{
  int i, j;
  for(j=0; j<4; ++j)
    for(i=0; i<4; ++i)
      M[j][i] = a[j][i] * k;
}

static inline void matrixR4x4f_scale_in_place(
  MatrixR4x4f& M,
  float k)
{
  int i, j;
  for(j=0; j<4; ++j)
    for(i=0; i<4; ++i)
      M[j][i] = M[j][i] * k;
}

static inline void matrixR4x4f_mul(
  MatrixR4x4f& M,
  const MatrixR4x4f& a,
  const MatrixR4x4f& b)
{
  MatrixR4x4f temp;
  int k, r, c;
  for(c=0; c<4; ++c) for(r=0; r<4; ++r) {
    temp[c][r] = 0.f;
    for(k=0; k<4; ++k)
      temp[c][r] += a[k][r] * b[c][k];
  }
  matrixR4x4f_dup(M, temp);
}

static inline void matrixR4x4f_mul_in_place(
  MatrixR4x4f& M,
  const MatrixR4x4f& a)
{
  MatrixR4x4f temp;
  int k, r, c;
  for(c=0; c<4; ++c) for(r=0; r<4; ++r) {
    temp[c][r] = 0.f;
    for(k=0; k<4; ++k)
      temp[c][r] += M[k][r] * a[c][k];
  }
  matrixR4x4f_dup(M, temp);
}

static inline void matrixR4x4f_mul_vectorR4(
  VectorR4f& r,
  const MatrixR4x4f& M,
  const VectorR4f& v)
{
  int i, j;
  for(j=0; j<4; ++j) {
    r[j] = 0.f;
    for(i=0; i<4; ++i)
      r[j] += M[i][j] * v[i];
  }
}

static inline void matrixR4x4f_mul_vectorR3(
  VectorR3f& r,
  const MatrixR4x4f& M,
  const VectorR3f& v)
{
  VectorR4f inTemp, outTemp;
  inTemp[0] = v[0];
  inTemp[1] = v[1];
  inTemp[2] = v[2];
  inTemp[3] = 1.0f;
  int i, j;
  for(j=0; j<4; ++j) {
    outTemp[j] = 0.f;
    for(i=0; i<4; ++i)
      outTemp[j] += M[i][j] * inTemp[i];
  }
  r[0] = outTemp[0];
  r[1] = outTemp[1];
  r[2] = outTemp[2];
}

static inline void matrixR4x4f_translate(
  MatrixR4x4f& T,
  float x,
  float y,
  float z)
{
  matrixR4x4f_identity(T);
  T[3][0] = x;
  T[3][1] = y;
  T[3][2] = z;
}

static inline void matrixR4x4f_translate_in_place(
  MatrixR4x4f& M,
  float x,
  float y,
  float z)
{
  VectorR4f t = {x, y, z, 0};
  VectorR4f r;
  int i;
  for (i = 0; i < 4; ++i)
  {
    matrixR4x4f_row(r, M, i);
    M[3][i] += (float)(dot(r, t));
  }
}

static inline void matrixR4x4f_translate_in_place(
  MatrixR4x4f& M,
  const VectorR3f& a)
{
  VectorR4f t = {a[X], a[Y], a[Z], 0.0f};
  VectorR4f r;
  int i;
  for (i = 0; i < 4; ++i)
  {
    matrixR4x4f_row(r, M, i);
    M[3][i] += (float)(dot(r, t));
  }
}

static inline void matrixR4x4f_translate_in_place(
  MatrixR4x4f& M,
  const VectorR2f& a)
{
  VectorR4f t = {a[X], a[Y], 0.0f, 0.0f};
  VectorR4f r;
  int i;
  for (i = 0; i < 4; ++i)
  {
    matrixR4x4f_row(r, M, i);
    M[3][i] += (float)(dot(r, t));
  }
}

static inline void matrixR4x4f_from_vectorR3_mul_outer(
  MatrixR4x4f& M,
  const VectorR3f& a,
  const VectorR3f& b)
{
  int i, j;
  for(i=0; i<4; ++i) for(j=0; j<4; ++j)
    M[i][j] = i<3 && j<3 ? a[i] * b[j] : 0.f;
}

static inline void matrixR4x4f_rotate(
  MatrixR4x4f& R,
  const MatrixR4x4f& M,
  float x,
  float y,
  float z,
  float angle)
{
  float s = sinf(angle);
  float c = cosf(angle);
  VectorR3f u = makeVector(x, y, z);
  if(length(u) > 1e-4)
  {
    MatrixR4x4f T, C, S = {{0}};
    normalize(u, 1.0e-6);
    matrixR4x4f_from_vectorR3_mul_outer(T, u, u);
    
    S[1][2] =  u[0];
    S[2][1] = -u[0];
    S[2][0] =  u[1];
    S[0][2] = -u[1];
    S[0][1] =  u[2];
    S[1][0] = -u[2];
    matrixR4x4f_scale(S, S, s);
    matrixR4x4f_identity(C);
    matrixR4x4f_sub(C, C, T);
    matrixR4x4f_scale(C, C, c);
    matrixR4x4f_add(T, T, C);
    matrixR4x4f_add(T, T, S);
    
    T[3][3] = 1.;
    matrixR4x4f_mul(R, M, T);
  } else {
    matrixR4x4f_dup(R, M);
  }
}

static inline void matrixR4x4f_rotate_X(
  MatrixR4x4f& Q,
  const MatrixR4x4f& M,
  float angle)
{
  float s = sinf(angle);
  float c = cosf(angle);
  MatrixR4x4f R = {
    {1.f, 0.f, 0.f, 0.f},
    {0.f,   c,   s, 0.f},
    {0.f,  -s,   c, 0.f},
    {0.f, 0.f, 0.f, 1.f}
  };
  matrixR4x4f_mul(Q, M, R);
}

static inline void matrixR4x4f_rotate_Y(
  MatrixR4x4f& Q,
  const MatrixR4x4f& M,
  float angle)
{
  float s = sinf(angle);
  float c = cosf(angle);
  MatrixR4x4f R = {
    {   c, 0.f,  -s, 0.f},
    { 0.f, 1.f, 0.f, 0.f},
    {   s, 0.f,   c, 0.f},
    { 0.f, 0.f, 0.f, 1.f}
  };
  matrixR4x4f_mul(Q, M, R);
}

static inline void matrixR4x4f_rotate_Z(
  MatrixR4x4f& Q,
  const MatrixR4x4f& M,
  float angle)
{
  float s = sinf(angle);
  float c = cosf(angle);
  MatrixR4x4f R = {
    {   c,   s, 0.f, 0.f},
    {  -s,   c, 0.f, 0.f},
    { 0.f, 0.f, 1.f, 0.f},
    { 0.f, 0.f, 0.f, 1.f}
  };
  matrixR4x4f_mul(Q, M, R);
}

static inline void matrixR4x4f_invert(
  MatrixR4x4f& T,
  const MatrixR4x4f& M)
{
  float idet;
  float s[6];
  float c[6];
  s[0] = M[0][0]*M[1][1] - M[1][0]*M[0][1];
  s[1] = M[0][0]*M[1][2] - M[1][0]*M[0][2];
  s[2] = M[0][0]*M[1][3] - M[1][0]*M[0][3];
  s[3] = M[0][1]*M[1][2] - M[1][1]*M[0][2];
  s[4] = M[0][1]*M[1][3] - M[1][1]*M[0][3];
  s[5] = M[0][2]*M[1][3] - M[1][2]*M[0][3];
  
  c[0] = M[2][0]*M[3][1] - M[3][0]*M[2][1];
  c[1] = M[2][0]*M[3][2] - M[3][0]*M[2][2];
  c[2] = M[2][0]*M[3][3] - M[3][0]*M[2][3];
  c[3] = M[2][1]*M[3][2] - M[3][1]*M[2][2];
  c[4] = M[2][1]*M[3][3] - M[3][1]*M[2][3];
  c[5] = M[2][2]*M[3][3] - M[3][2]*M[2][3];
  
  /* Assumes it is invertible */
  idet = 1.0f/( s[0]*c[5]-s[1]*c[4]+s[2]*c[3]+s[3]*c[2]-s[4]*c[1]+s[5]*c[0] );
  
  T[0][0] = ( M[1][1] * c[5] - M[1][2] * c[4] + M[1][3] * c[3]) * idet;
  T[0][1] = (-M[0][1] * c[5] + M[0][2] * c[4] - M[0][3] * c[3]) * idet;
  T[0][2] = ( M[3][1] * s[5] - M[3][2] * s[4] + M[3][3] * s[3]) * idet;
  T[0][3] = (-M[2][1] * s[5] + M[2][2] * s[4] - M[2][3] * s[3]) * idet;
  
  T[1][0] = (-M[1][0] * c[5] + M[1][2] * c[2] - M[1][3] * c[1]) * idet;
  T[1][1] = ( M[0][0] * c[5] - M[0][2] * c[2] + M[0][3] * c[1]) * idet;
  T[1][2] = (-M[3][0] * s[5] + M[3][2] * s[2] - M[3][3] * s[1]) * idet;
  T[1][3] = ( M[2][0] * s[5] - M[2][2] * s[2] + M[2][3] * s[1]) * idet;
  
  T[2][0] = ( M[1][0] * c[4] - M[1][1] * c[2] + M[1][3] * c[0]) * idet;
  T[2][1] = (-M[0][0] * c[4] + M[0][1] * c[2] - M[0][3] * c[0]) * idet;
  T[2][2] = ( M[3][0] * s[4] - M[3][1] * s[2] + M[3][3] * s[0]) * idet;
  T[2][3] = (-M[2][0] * s[4] + M[2][1] * s[2] - M[2][3] * s[0]) * idet;
  
  T[3][0] = (-M[1][0] * c[3] + M[1][1] * c[1] - M[1][2] * c[0]) * idet;
  T[3][1] = ( M[0][0] * c[3] - M[0][1] * c[1] + M[0][2] * c[0]) * idet;
  T[3][2] = (-M[3][0] * s[3] + M[3][1] * s[1] - M[3][2] * s[0]) * idet;
  T[3][3] = ( M[2][0] * s[3] - M[2][1] * s[1] + M[2][2] * s[0]) * idet;
}

#ifdef ADVANCED_MATRIX_FUNCTIONS
static inline void matrixR4x4f_orthonormalize(
  MatrixR4x4f& R,
  const MatrixR4x4f& M)
{
  float s = 1.;
  VectorR3f h;
  
  matrixR4x4f_dup(R, M);
  vectorR3_norm(R[2], R[2]);
  
  s = vectorR3_mul_inner(R[1], R[2]);
  vectorR3_scale(h, R[2], s);
  vectorR3_sub(R[1], R[1], h);
  vectorR3_norm(R[2], R[2]);
  
  s = vectorR3_mul_inner(R[1], R[2]);
  vectorR3_scale(h, R[2], s);
  vectorR3_sub(R[1], R[1], h);
  vectorR3_norm(R[1], R[1]);
  
  s = vectorR3_mul_inner(R[0], R[1]);
  vectorR3_scale(h, R[1], s);
  vectorR3_sub(R[0], R[0], h);
  vectorR3_norm(R[0], R[0]);
}
#endif

static inline void matrixR4x4f_frustum(
  MatrixR4x4f& M,
  float l,
  float r,
  float b,
  float t,
  float n,
  float f)
{
  M[0][0] = 2.f*n/(r-l);
  M[0][1] = M[0][2] = M[0][3] = 0.f;
  
  M[1][1] = 2.f*n/(t-b);
  M[1][0] = M[1][2] = M[1][3] = 0.f;
  
  M[2][0] = (r+l)/(r-l);
  M[2][1] = (t+b)/(t-b);
  M[2][2] = -(f+n)/(f-n);
  M[2][3] = -1.f;
  
  M[3][2] = -2.f*(f*n)/(f-n);
  M[3][0] = M[3][1] = M[3][3] = 0.f;
}

static inline void matrixR4x4f_ortho(
  MatrixR4x4f& M,
  float l,
  float r,
  float b,
  float t,
  float n,
  float f)
{
  M[0][0] = 2.f/(r-l);
  M[0][1] = M[0][2] = M[0][3] = 0.f;
  
  M[1][1] = 2.f/(t-b);
  M[1][0] = M[1][2] = M[1][3] = 0.f;
  
  M[2][2] = -2.f/(f-n);
  M[2][0] = M[2][1] = M[2][3] = 0.f;
  
  M[3][0] = -(r+l)/(r-l);
  M[3][1] = -(t+b)/(t-b);
  M[3][2] = -(f+n)/(f-n);
  M[3][3] = 1.f;
}

static inline void matrixR4x4f_perspective(
  MatrixR4x4f& m,
  float y_fov,
  float aspect,
  float n,
  float f)
{
  /* NOTE: Degrees are an unhandy unit to work with.
   * linmath.h uses radians for everything! */
  float const a = 1.f / (float) tan(y_fov / 2.f);
  
  m[0][0] = a / aspect;
  m[0][1] = 0.f;
  m[0][2] = 0.f;
  m[0][3] = 0.f;
  
  m[1][0] = 0.f;
  m[1][1] = a;
  m[1][2] = 0.f;
  m[1][3] = 0.f;
  
  m[2][0] = 0.f;
  m[2][1] = 0.f;
  m[2][2] = -((f + n) / (f - n));
  m[2][3] = -1.f;
  
  m[3][0] = 0.f;
  m[3][1] = 0.f;
  m[3][2] = -((2.f * f * n) / (f - n));
  m[3][3] = 0.f;
}

static inline void matrixR4x4f_look_at(
  MatrixR4x4f& m,
  const VectorR3f& eye,
  const VectorR3f& center,
  const VectorR3f& up)
{
  /* Adapted from Android's OpenGL Matrix.java.                        */
  /* See the OpenGL GLUT documentation for gluLookAt for a description */
  /* of the algorithm. We implement it in a straightforward way:       */
  
  /* TODO: The negation of of can be spared by swapping the order of
   *       operands in the following cross products in the right way. */
  VectorR3f f;
  VectorR3f s;
  VectorR3f t;
  
  sub(center, eye, f);
  normalize(f, 1.0e-6);
  
  cross(f, up, s);
  normalize(s, 1.0e-6);
  
  cross(s, f, t);
  
  m[0][0] =  s[0];
  m[0][1] =  t[0];
  m[0][2] = -f[0];
  m[0][3] =   0.f;
  
  m[1][0] =  s[1];
  m[1][1] =  t[1];
  m[1][2] = -f[1];
  m[1][3] =   0.f;
  
  m[2][0] =  s[2];
  m[2][1] =  t[2];
  m[2][2] = -f[2];
  m[2][3] =   0.f;
  
  m[3][0] =  0.f;
  m[3][1] =  0.f;
  m[3][2] =  0.f;
  m[3][3] =  1.f;
  
  matrixR4x4f_translate_in_place(m, -eye[0], -eye[1], -eye[2]);
}

#ifdef ADVANCED_MATRIX_FUNCTIONS
typedef float Quaternion[4];

static inline void quaternion_identity(
  Quaternion& q)
{
  q[0] = q[1] = q[2] = 0.f;
  q[3] = 1.f;
}

static inline void quaternion_add(
  Quaternion& r,
  const Quaternion& a,
  const Quaternion& b)
{
  int i;
  for(i=0; i<4; ++i)
    r[i] = a[i] + b[i];
}

static inline void quaternion_sub(
  Quaternion& r,
  const Quaternion& a,
  const Quaternion& b)
{
  int i;
  for(i=0; i<4; ++i)
    r[i] = a[i] - b[i];
}

static inline void quaternion_mul(
  Quaternion& r,
  const Quaternion& p,
  const Quaternion& q)
{
  VectorR3f w;
  cross(p, q, r);
  vectorR3_scale(w, p, q[3]);
  vectorR3_add(r, r, w);
  vectorR3_scale(w, q, p[3]);
  vectorR3_add(r, r, w);
  r[3] = p[3]*q[3] - vectorR3_mul_inner(p, q);
}

static inline void quaternion_scale(
  Quaternion& r,
  const Quaternion& v,
  float s)
{
  int i;
  for(i=0; i<4; ++i)
    r[i] = v[i] * s;
}

static inline float quaternion_inner_product(
  const Quaternion& a,
  const Quaternion& b)
{
  float p = 0.f;
  int i;
  for(i=0; i<4; ++i)
    p += b[i]*a[i];
  return p;
}

static inline void quaternion_conj(
  Quaternion& r,
  const Quaternion& q)
{
  int i;
  for(i=0; i<3; ++i)
    r[i] = -q[i];
  r[3] = q[3];
}

static inline void quaternion_rotate(
  Quaternion& r,
  float angle,
  const VectorR3& axis)
{
  int i;
  VectorR3 v;
  vectorR3_scale(v, axis, sinf(angle / 2));
  for(i=0; i<3; ++i)
    r[i] = v[i];
  r[3] = cosf(angle / 2);
}

#define quaternion_norm vectorR4_norm
static inline void quaternion_mul_vectorR3(
  VectorR3 r,
  const Quaternion& q,
  VectorR3 v)
{
  /*
   * Method by Fabian 'ryg' Giessen (of Farbrausch)
   t = 2 * cross(q.xyz, v)
   v' = v + q.w * t + cross(q.xyz, t)
   */
  VectorR3 t = {q[0], q[1], q[2]};
  VectorR3 u = {q[0], q[1], q[2]};
  
  vectorR3_mul_cross(t, t, v);
  vectorR3_scale(t, t, 2);
  
  vectorR3_mul_cross(u, u, t);
  vectorR3_scale(t, t, q[3]);
  
  vectorR3_add(r, v, t);
  vectorR3_add(r, r, u);
}

static inline void matrixR4x4f_from_quaternion(
  MatrixR4x4f& M,
  const Quaternion& q)
{
  float a = q[3];
  float b = q[0];
  float c = q[1];
  float d = q[2];
  float a2 = a*a;
  float b2 = b*b;
  float c2 = c*c;
  float d2 = d*d;
  
  M[0][0] = a2 + b2 - c2 - d2;
  M[0][1] = 2.f*(b*c + a*d);
  M[0][2] = 2.f*(b*d - a*c);
  M[0][3] = 0.f;
  
  M[1][0] = 2*(b*c - a*d);
  M[1][1] = a2 - b2 + c2 - d2;
  M[1][2] = 2.f*(c*d + a*b);
  M[1][3] = 0.f;
  
  M[2][0] = 2.f*(b*d + a*c);
  M[2][1] = 2.f*(c*d - a*b);
  M[2][2] = a2 - b2 - c2 + d2;
  M[2][3] = 0.f;
  
  M[3][0] = M[3][1] = M[3][2] = 0.f;
  M[3][3] = 1.f;
}

static inline void matrixR4x4fo_mul_quaternion(
  MatrixR4x4f& R,
  MatrixR4x4f& M,
  const Quaternion& q)
{
  /*  XXX: The way this is written only works for othogonal matrices. */
  /* TODO: Take care of non-orthogonal case. */
  quaternion_mul_vectorR3(R[0], q, M[0]);
  quaternion_mul_vectorR3(R[1], q, M[1]);
  quaternion_mul_vectorR3(R[2], q, M[2]);
  
  R[3][0] = R[3][1] = R[3][2] = 0.f;
  R[3][3] = 1.f;
}

static inline void quaternion_from_matrixR4x4f(
  Quaternion& q,
  const MatrixR4x4f& M)
{
  float r=0.f;
  int i;
  
  int perm[] = { 0, 1, 2, 0, 1 };
  int *p = perm;
  
  for(i = 0; i<3; i++) {
    float m = M[i][i];
    if( m < r )
      continue;
    m = r;
    p = &perm[i];
  }
  
  r = (float) sqrt(1.f + M[p[0]][p[0]] - M[p[1]][p[1]] - M[p[2]][p[2]] );
  
  if(r < 1e-6) {
    q[0] = 1.f;
    q[1] = q[2] = q[3] = 0.f;
    return;
  }
  
  q[0] = r/2.f;
  q[1] = (M[p[0]][p[1]] - M[p[1]][p[0]])/(2.f*r);
  q[2] = (M[p[2]][p[0]] - M[p[0]][p[2]])/(2.f*r);
  q[3] = (M[p[2]][p[1]] - M[p[1]][p[2]])/(2.f*r);
}
#endif

static inline std::string matrixR4x4f_info(
  const MatrixR4x4f& M)
{
  char numberString[100];
  std::string info;
  info = "  [";
  for (unsigned int c = 0; c < 4; ++c)
  {
    for (unsigned int r = 0; r < 4; ++r)
    {
      sprintf(numberString, "%12.4f", M[r][c]);
      info += numberString;
    }
    if (c != 3) info += " |\n   ";
    else info +=  " ]\n";
  }
  return info;
}

/** 
 * Assigns an orthographic projetion to the input matrix based on the input parameters
 *
 * @ingroup elements
 */
inline void setOrthograhpicProjection(
  MatrixR4x4f& matrix,
  float width,
  float height,
  float ratio,
  float nearRange,
  float farRange)
{
  // Determine the new view height
  matrixR4x4f_identity(matrix);
  float preferredAspectRatio = width / height;
  float viewportAspectRatio = ratio;
  float viewHeight;
  float zoom = 1.0f;
  if (viewportAspectRatio < 1.0f)
  {
    if (preferredAspectRatio < viewportAspectRatio)
    {
      viewHeight = height;
    }
    else
    {
      viewHeight = width / viewportAspectRatio;
    }
  }
  else
  {
    if (preferredAspectRatio > viewportAspectRatio)
    {
      viewHeight = width / viewportAspectRatio;
    }
    else
    {
      viewHeight = height;
    }
  }
  float left = -viewHeight * zoom / 2.0f * ratio;
  float right = viewHeight * zoom / 2.0f * ratio;
  float bottom = -viewHeight * zoom / 2.0f;
  float top = viewHeight * zoom / 2.0f;
  float nearMinusFar = nearRange - farRange;
  matrixR4x4f_assign(matrix,
    2.0f / (right - left), 0.0f, 0.0f, -(right + left) / (right - left),
    0.0f, 2.0f / (top - bottom), 0.0f, -(top + bottom) / (top - bottom),
    0.0f, 0.0f, 2.0f / nearMinusFar, nearMinusFar / nearMinusFar,
    0.0f, 0.0f, 0.0f, 1.0f);
}

//-------------------------
// New vector class support
//-------------------------
static inline void matrixR4x4f_scale(
  MatrixR4x4f& matrix,
  const VectorR3f& scaling)
{
  MatrixR4x4f scaleMatrix;
  matrixR4x4f_identity(scaleMatrix);
  scaleMatrix[0][0] = scaling[X];
  scaleMatrix[1][1] = scaling[Y];
  scaleMatrix[2][2] = scaling[Z];
  scaleMatrix[3][3] = 1.0f;
  matrixR4x4f_mul_in_place(matrix, scaleMatrix);
}

static inline void matrixR4x4f_scale(
  MatrixR4x4f& matrix,
  const VectorR2f& scaling)
{
  MatrixR4x4f scaleMatrix;
  matrixR4x4f_identity(scaleMatrix);
  scaleMatrix[0][0] = scaling[X];
  scaleMatrix[1][1] = scaling[Y];
  scaleMatrix[2][2] = 1.0f;
  scaleMatrix[3][3] = 1.0f;
  matrixR4x4f_mul_in_place(matrix, scaleMatrix);
}

static inline void matrixR4x4f_reflect_X(
  MatrixR4x4f& matrix)
{
  MatrixR4x4f scaleMatrix;
  matrixR4x4f_identity(scaleMatrix);
  scaleMatrix[0][0] = 1.0f;
  scaleMatrix[1][1] = -1.0f;
  scaleMatrix[2][2] = 1.0f;
  scaleMatrix[3][3] = 1.0f;
  matrixR4x4f_mul_in_place(matrix, scaleMatrix);
}

static inline void matrixR4x4f_reflect_Y(
  MatrixR4x4f& matrix)
{
  MatrixR4x4f scaleMatrix;
  matrixR4x4f_identity(scaleMatrix);
  scaleMatrix[0][0] = -1.0f;
  scaleMatrix[1][1] = 1.0f;
  scaleMatrix[2][2] = 1.0f;
  scaleMatrix[3][3] = 1.0f;
  matrixR4x4f_mul_in_place(matrix, scaleMatrix);
}

static inline void transform(
  const MatrixR4x4f& M,
  const VectorR2f& v,
  VectorR2f& result)
{
  VectorR4f v0;
  v0[0] = v[0];
  v0[1] = v[1];
  v0[2] = 0.0f;
  v0[3] = 1.0f;

  VectorR4f res;
  for (unsigned int r = 0; r < 4; ++r)
  {
    res[r] = 0.0f;
    for (unsigned int c = 0; c < 4; ++c) res[r] += M[c][r] * v0[c];
  }

  result[0] = res[0];
  result[1] = res[1];
}

static inline void transform(
  const MatrixR4x4f& M,
  const VectorR3f& v,
  VectorR2f& result)
{
  VectorR4f v0;
  v0[0] = v[0];
  v0[1] = v[1];
  v0[2] = v[2];
  v0[3] = 1.0f;

  VectorR4f res;
  for (unsigned int r = 0; r < 4; ++r)
  {
    res[r] = 0.0f;
    for (unsigned int c = 0; c < 4; ++c) res[r] += M[c][r] * v0[c];
  }

  result[0] = res[0];
  result[1] = res[1];
}

static inline void transform(
  const MatrixR4x4f& M,
  const VectorR2f& v,
  VectorR3f& result)
{
  VectorR4f v0;
  v0[0] = v[0];
  v0[1] = v[1];
  v0[2] = 0.0f;
  v0[3] = 1.0f;

  VectorR4f res;
  for (unsigned int r = 0; r < 4; ++r)
  {
    res[r] = 0.0f;
    for (unsigned int c = 0; c < 4; ++c) res[r] += M[c][r] * v0[c];
  }

  result[0] = res[0];
  result[1] = res[1];
  result[2] = res[3];
}

static inline void transform(
  const MatrixR4x4f& M,
  const VectorR3f& v,
  VectorR3f& result)
{
  VectorR4f v0;
  v0[0] = v[0];
  v0[1] = v[1];
  v0[2] = v[2];
  v0[3] = 1.0f;

  VectorR4f res;
  for (unsigned int r = 0; r < 4; ++r)
  {
    res[r] = 0.0f;
    for (unsigned int c = 0; c < 4; ++c) res[r] += M[c][r] * v0[c];
  }

  result[0] = res[0];
  result[1] = res[1];
  result[2] = res[2];
}

static inline void transformVector(
  const MatrixR4x4f& t,
  const VectorR3f& v,
  VectorR3f& result)
{
  VectorR3f v0, v1;
  v0[X] = 0.0;
  v0[Y] = 0.0;
  v0[Z] = 0.0;
  transform(t, v0, v1);
  VectorR3f v2, v3;
  v2[X] = v[X];
  v2[Y] = v[Y];
  v2[Z] = v[Z];

  transform(t, v2, v3);
  result[X] = v3[X] - v1[X];
  result[Y] = v3[Y] - v1[Y];
  result[Z] = v3[Z] - v1[Z];
}

} // namespace pdp

