/** @file
 * 
 * Time sequence classes and functions
 * 
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <boost/property_tree/ptree.hpp>
#include <boost/property_tree/json_parser.hpp>
#include <utilities/stringUtils.h>
#include <utilities/timeUtils.h>
#include <elements/export.h>

namespace pdp {

/**
 * Object for holding data for a singe time step for sequence data.
 *
 * @ingroup elements
 */
class PDPELEMENTS_EXPORT TimeSample
{
public:
  TimeSample()
  : m_offset(0.0),
    m_step(0.0),
    m_temp(0.0f),
    m_battery(0.0f),
    m_humidity(0.0f),
    m_rssi(0.0f),
    m_rollingHumidity(0.0f)
  {}

  ~TimeSample() {}
  
  std::chrono::system_clock::time_point timePoint() const
  { return m_tp; }
  
  bool operator<(const TimeSample &rhs) const {return m_tp < rhs.m_tp;}

public:
  std::chrono::system_clock::time_point m_tp;  
  double m_offset;
  double m_step;
  float m_temp;
  float m_battery;
  float m_humidity;
  float m_rssi;
  float m_rollingHumidity;
};

/**
 * Object for holding time sequence data.
 *
 * @ingroup elements
 */
class TimeSampleSequence
{
public:
  enum ValueType
  {
    ValueTypeTemp = 0,
    ValueTypeBattery,
    ValueTypeHumidity,
    ValueTypeRssi,
    ValueTypeRollingHumidity
  };

  enum StatusType
  {
    StatusTypeUnknown = 0,
    StatusTypeFailure,
    StatusTypeNonFailure
  };

  TimeSampleSequence()
  : m_minTemp(0.0),
    m_avgTemp(0.0),
    m_maxTemp(0.0),
    m_minBattery(0.0),
    m_avgBattery(0.0),
    m_maxBattery(0.0),
    m_minHumidity(0.0),
    m_avgHumidity(0.0),
    m_maxHumidity(0.0),
    m_minRssi(0.0),
    m_avgRssi(0.0),
    m_maxRssi(0.0),
    m_minRollingHumidity(0.0),
    m_avgRollingHumidity(0.0),
    m_maxRollingHumidity(0.0),
    m_numberOfOriginalSamples(0),
    m_span(0.0),
    m_minimumStep(0.0),
    m_maximumStep(0.0),
    m_truthStatus(StatusTypeUnknown),
    m_classifyStatus(StatusTypeUnknown),
    m_confidence(0.0f)
  {}

  ~TimeSampleSequence() {}

  void loadFromPropertyTree(
    const boost::property_tree::ptree& pt);

  void finalizeAfterDataLoad();

  std::string info(
    bool details) const;

  void clear();

  void addSample(
    const TimeSample& timeSample)
  { m_samples.push_back(timeSample); }

  void getInterpolatedSampleValues(
    TimeSample& sample) const;

  void getComponentSampleValues(
    double timeStep,
    TimeSample& sample) const;

  void computeTimeParameters();

  int numberOfSamples() const
  { return (int)(m_samples.size()); }

  const std::vector<TimeSample>& samples() const
  { return m_samples; }

  int numberOfOriginalSamples() const
  { return m_numberOfOriginalSamples; }

  void setNumberOfOriginalSamples(
    int numberOfOriginalSamples)
  { m_numberOfOriginalSamples = numberOfOriginalSamples; }

  double span() const
  { return m_span; }
  
  double minimumStep() const
  { return m_minimumStep; }
  
  double maximumStep() const
  { return m_maximumStep; }

  void setTruthStatus(
    StatusType status)
  { m_truthStatus = status; }

  StatusType truthStatus()
  { return m_truthStatus; }

  void setClassificationStatus(
    StatusType status,
    float confidence)
  { m_classifyStatus = status; m_confidence = confidence; }

  StatusType classificationStatus()
  { return m_classifyStatus; }

  float confidence()
  { return m_confidence; }

  const std::string& id() const
  { return m_id; }

  void setId(
    const std::string& id)
  { m_id = id; }

  const std::string& name() const
  { return m_name; }

  void setName(
    const std::string& name)
  { m_name = name; }

  const std::string& unitId() const
  { return m_unitId; }

  void setUnitId(
    const std::string& unitId)
  { m_unitId = unitId; }

  float minTemp() const
  { return m_minTemp; }

  float avgTemp() const
  { return m_avgTemp; }

  float maxTemp() const
  { return m_maxTemp; }
  
  void getSequenceAndLabel(
    size_t& imagesOffset,
    float* images,
    size_t& labelsOffset,
    float* labels);

  void computeStats();      

  void updateStats(
    float& minTemp,
    float& maxTemp,
    float& minBattery,
    float& maxBattery,
    float& minHumidity,
    float& maxHumidity,
    float& minRssi,
    float& maxRssi,
    float& minRollingHumidity,
    float& maxRollingHumidity);      

  void getRanges(
    float& tempRange,
    float& batteryRange,
    float& humidityRange,
    float& rssiRange,
    float& rollingHumidityRange);

  void updateRanges(
    float& tempRange,
    float& batteryRange,
    float& humidityRange,
    float& rssiRange,
    float& rollingHumidityRange);
    
  void normalize(
    float minTemp,
    float maxTemp,
    float minBattery,
    float maxBattery,
    float minHumidity,
    float maxHumidity,
    float minRssi,
    float maxRssi,
    float minRollingHumidity,
    float maxRollingHumidity);      

  void normalizeRange(
    float tempRange,
    float batteryRange,
    float humidityRange,
    float rssiRange,
    float rHumidityRange);
    
  void normalizeLocal();

  void sampleValue(
    ValueType valueType,
    int sampleIdx,
    float& x,
    float& y);

private:

private:
  std::string m_id;
  std::string m_name;
  std::string m_unitId;
  std::vector<TimeSample> m_samples;
  float m_minTemp;
  float m_avgTemp;
  float m_maxTemp;
  float m_minBattery;
  float m_avgBattery;
  float m_maxBattery;
  float m_minHumidity;
  float m_avgHumidity;
  float m_maxHumidity;
  float m_minRssi;
  float m_avgRssi;
  float m_maxRssi;
  float m_minRollingHumidity;
  float m_avgRollingHumidity;
  float m_maxRollingHumidity;
  int m_numberOfOriginalSamples;
  double m_span;
  double m_minimumStep;
  double m_maximumStep;
  StatusType m_truthStatus;
  StatusType m_classifyStatus;
  float m_confidence;
};

/**
 * Clears the results seqeunce and fills it with data from the input sequence with uniform time steps. Uses linear
 * interpolation to sample the input sequence
 *
 * @ingroup elements
 */
PDPELEMENTS_EXPORT void createUniformTimeSampleSequence(
  int sequenceIdx,
  const TimeSampleSequence& inputSequence,
  double timeSpan,
  int numberOfTimeSteps,
  bool failed,
  TimeSampleSequence& resultSequence,
  std::string& info);

} // namespace pdp



