/** @file
 * 
 * Multidimensional triangle lists 
 * 
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <vector>
#include <elements/export.h>
#include <elements/Color.h>
#include <elements/VectorR.h>
#include <elements/TriangleR.h>

namespace pdp {

/**
 * structure to define light properties
 */
struct Light
{
  pdp::Color color;
  pdp::VectorR4f ambient;
  pdp::VectorR4f diffuse;
  pdp::VectorR4f specular;
  pdp::VectorR4f position;
  float strength;
};

/**
 * structure to define material properties
 */
struct Material
{
  pdp::VectorR4f ambient;
  pdp::VectorR4f diffuse;
  pdp::VectorR4f specular;
  pdp::VectorR4f emission;
  float shininess;
};

inline void putMaterial(
  const Material& m,
  std::ostream& outputStream)
{
  putVectorR(m.ambient, outputStream);
  putVectorR(m.diffuse, outputStream);
  putVectorR(m.specular, outputStream);
  putVectorR(m.emission, outputStream);
  putR<float>(m.shininess, outputStream);
}

inline void getMaterial(
  std::istream& inputStream,
  Material& m)
{
  getVectorR(inputStream, m.ambient);
  getVectorR(inputStream, m.diffuse);
  getVectorR(inputStream, m.specular);
  getVectorR(inputStream, m.emission);
  m.shininess = getR<float>(inputStream);
}

/**
 * Represents a multidimensional triangle list. The list is of
 * disconnected triangles generated by algorithms such as marching
 * cubes.
 *
 * @ingroup pdp_elements
 */
template<unsigned int N, typename E>
class TriangleListR
{
public:
  /**
   * storage structure for triangles
   */
  struct Triangle
  {
    unsigned int v0;
    unsigned int v1;
    unsigned int v2;
  };

  /**
   * storage structure index reference lists
   */
  struct IndexList
  {
    std::vector<unsigned int> indexes;
  };

  /**
   * Constructs an empty triangle list. This initial list contains no
   * triangles.
   */
  TriangleListR() {}

  /**
   * Constructs a new triangle list and copies the settings from the passed-in triangle list object.
   */
  TriangleListR(
    const TriangleListR& triangleList)
  {
    m_positions = triangleList.m_positions;
    m_normals = triangleList.m_normals;
    m_triangles = triangleList.m_triangles;
  }

  /**
   * Returns a const pointer to an array containing the positions of the internal triangle list vertices.
   */
  const E* positionsPointer() const
  {
		if (m_positions.size() == 0)
    	return 0;
    return m_positions[0].elements();
  }

  /**
   * Returns a const pointer to an array containing the normals of the internal triangle list.
   */
  const E* normalsPointer() const
  {
		if (m_normals.size() == 0)
    	return 0;
    return m_normals[0].elements();
  }

  /**
   * Returns a const reference to the internal vertex list.
   */
  const std::vector< VectorR<N, E> >& internalVertexList() const
  {
    return m_positions;
  }

  /**
   * Returns a const reference to the internal normals list.
   */
  const std::vector< VectorR<N, E> >& internalNormalList() const
  {
    return m_normals;
  }

  /**
   * Returns a const reference to the internal triangle list.
   */
  const std::vector<Triangle>& internalTriangleList() const
  {
    return m_triangles;
  }

  /**
   * Adds @a vertex to the end of the triangle vertex list. This method
   * simply adds the vertex to the end of the list and does not check
   * if the input vertex is already in the table.
   */
  void addVertex(
    bool addProcessingElements,
    const VectorR<N, E>& p,
    const VectorR<N, E>& n);

  /**
   * Returns the number of vertexes.
   */
  size_t numVertexes() const
  { return m_positions.size(); }

  /**
   * Returns the vertex at @a idx, in model coordinates.
   */
  void vertex(
    unsigned int idx,
    VectorR<N, E>& p,
    VectorR<N, E>& n) const;

  /**
   * Adds a triangle to the triangle list. This method requires that a
   * valid vertex index (i.e. the vertex is already in the vertex table)
   * be supplied for each vertex of the triangle.
   */
  void addTriangle(
    bool addProcessingElements,
    unsigned int v0Idx,
    unsigned int v1Idx,
    unsigned int v2Idx);

  /**
   * Returns the triangle vertexes at @a idx, in model coordinates.
   */
  void triangle(
    unsigned int idx,
    VectorR<N, E>& p0,
    VectorR<N, E>& n0,
    VectorR<N, E>& p1,
    VectorR<N, E>& n1,
    VectorR<N, E>& p2,
    VectorR<N, E>& n2) const;
 
  /**
   * Returns the number of triangles.
   */
  unsigned int numTriangles() const
  { return static_cast<unsigned int>(m_triangles.size()); }
  
  /**
   * Clears the triangle list.
   */
  void clear()
  {
    m_positions.clear();
    m_normals.clear();
    m_triangles.clear();
    m_triangleLists.clear();
    m_neighborVertexLists.clear();
    m_triangleNormals.clear();
  }
  
  /**
   * Clears the processing elements.
   */
  void clearProcessingElements()
  {
    m_triangleLists.clear();
    m_neighborVertexLists.clear();
    m_triangleNormals.clear();
  }

  /**
   * Smooths the trianlge list using a parameteric low pass filter. This method requires
   * that the vertexes neighbor list is available
   */
  void lowpass(
    double lamda,
    double sigma,
    unsigned int numberOfIterations);

  /**
   * Computes the normal for each vertex of the surface. This method requires
   * that the vertexes triangle list is available
   */
  void computeNormals();

  /**
   * This methods reorders the vertexes and normals so that the memory array
   * can be used directly for rendering
   */
	void reorderVertexes()
  {
		// Build the new normals and positions lists
  	VectorR<N, E> p0, n0, p1, n1, p2, n2;
  	std::vector< VectorR<N, E> > positionsFlat;
  	std::vector< VectorR<N, E> > normalsFlat;
    for (unsigned int i = 0; i < m_triangles.size(); i++)
    {
			// Get the positions and normals for the current triangle
      triangle(i, p0, n0, p1, n1, p2, n2);

			// Add the current triangle vertices and normals to the lists in display order
			positionsFlat.push_back(p0);
			positionsFlat.push_back(p1);
			positionsFlat.push_back(p2);
			normalsFlat.push_back(n0);
			normalsFlat.push_back(n1);
			normalsFlat.push_back(n2);

			// Update the vertex ids for the current triangle
      m_triangles[i].v0 = i*3;
      m_triangles[i].v1 = i*3 + 1;
      m_triangles[i].v2 = i*3 + 2;
    }

		// Transfer the new list to permenent storage
		m_positions.clear();
		m_normals.clear();
    for (unsigned int i = 0; i < positionsFlat.size(); i++)
    {
			m_positions.push_back(positionsFlat[i]);
			m_normals.push_back(normalsFlat[i]);    	
    }
		positionsFlat.clear();
		normalsFlat.clear();    
  }

  //---------------------------------------------------------------------------
  // Friend functions
  //--------------------------------------------------------------------------- 
  /**
   * The method gets the next TriangleListR from the input stream and
   * decodes it from little endian format.
   */
  friend void getTriangleListR(
    std::istream& inputStream,
    TriangleListR<N, E>& t)
  {
    // Get the vertex table
    auto numVertexes = getR<unsigned int>(inputStream);
    t.m_positions.clear();
    t.m_normals.clear();
    t.m_triangleLists.clear();
    t.m_neighborVertexLists.clear();
    VectorR<N, E> position;
    VectorR<N, E> normal;
    
    // pre-sizing the vector for positions and normals to prevent copying
    t.m_positions.reserve(numVertexes);
    t.m_normals.reserve(numVertexes);
    for (unsigned int i = 0; i < numVertexes; i++)
    {
      getVectorR(inputStream, position);
      getVectorR(inputStream, normal);
      t.m_positions.push_back(position);
      t.m_normals.push_back(normal);
    }

    // Get the triangle table
    auto numTriangles = getR<unsigned int>(inputStream);
    t.m_triangles.clear();
    Triangle triangle;
    
    // pre-sizing the vectors for triangles to prevent copying
    t.m_triangles.reserve(numVertexes);
    for (unsigned int i = 0; i < numTriangles; i++)
    {
      triangle.v0 = getR<unsigned int>(inputStream);
      triangle.v1 = getR<unsigned int>(inputStream);
      triangle.v2 = getR<unsigned int>(inputStream);
      t.m_triangles.push_back(triangle);
    }

		// Reorder the vertexes for display. This is done to support old format
    // triagle lists which may not have been reordered on creation.
		t.reorderVertexes();
    t.m_positions.shrink_to_fit();
    t.m_normals.shrink_to_fit();
    t.m_triangles.shrink_to_fit();
  }

  /**
   * The method encodes the input TriangleListR into little endian format
   * and writes it to the output stream.
   */
  friend void putTriangleListR(
    const TriangleListR<N, E>& t,
    std::ostream& outputStream)
  {
    // Write the vertexes table
    putR<unsigned int>( static_cast<unsigned int>(t.m_positions.size()), outputStream);
    VectorR<N, E> temp;
    for (unsigned int i = 0; i < t.m_positions.size(); i++)
    {
      putVectorR(t.m_positions[i], outputStream);
      putVectorR(t.m_normals[i], outputStream);
    }

    // Write the triangles table
    putR<unsigned int>( static_cast<unsigned int>(t.m_triangles.size()), outputStream );
    for (unsigned int i = 0; i < t.m_triangles.size(); i++)
    {
      putR<unsigned int>(t.m_triangles[i].v0, outputStream);
      putR<unsigned int>(t.m_triangles[i].v1, outputStream);
      putR<unsigned int>(t.m_triangles[i].v2, outputStream);
    }
  }

private:
  // State information saved to binary database file
  std::vector<VectorR<N, E>> m_positions;
  std::vector<VectorR<N, E>> m_normals;
  std::vector<Triangle> m_triangles;

  // *** Dynamic information not saved to binary database file
  std::vector<IndexList> m_triangleLists;
  std::vector<IndexList> m_neighborVertexLists;
  std::vector<VectorR<N, E>> m_triangleNormals;
};

/*
 * Short type names for common triangle list specializations.
 */
typedef TriangleListR<2, int> TriangleListR2i;
typedef TriangleListR<2, float> TriangleListR2f;
typedef TriangleListR<2, double> TriangleListR2d;
typedef TriangleListR<3, int> TriangleListR3i;
typedef TriangleListR<3, float> TriangleListR3f;
typedef TriangleListR<3, double> TriangleListR3d;

} // namespace pdp

#include <elements/TriangleListR.t>

