/** @file
 *
 * Triangles composed of vectors with arbitrary element type.
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <assert.h>
#include <ostream>
#include <elements/export.h>
#include <elements/VectorR.h>

namespace pdp
{
/**
 * The TriangleR template class represents triangles composed of
 * fixed-size, real vectors.  The dimension and element type of the triangle
 * are specified as part of the type, and are thus fixed at compile
 * time.
 *
 * @a N is the number of elements, which must be greater than zero.
 * @a E is the element type, typically @c double, @c float, or @c int.
 *
 * @ingroup pdp_elements
 */
template<unsigned int N, typename E>
class TriangleR
{
public:
  /**
   * The number of dimensions in line segments of this type.
   */
  static const unsigned int SIZE = N;

  /**
   * Constructs a triangle whose elements are uninitialized.
   */
  inline TriangleR() {}

  /**
   * Destroys a triangle.
   */
  inline ~TriangleR() {}

  /**
   * Constructs a triangle based on the input vectors @a p0, @a p1, @a p2
   * and the normal vectors  @a n0, @a n1, @a n2. The type of the input vectors
   * can be different from that of the result line segment as long as the types
   * are assignment compatible.
   */
  TriangleR(
    const VectorR<N, E>& p0,
    const VectorR<N, E>& p1,
    const VectorR<N, E>& p2,
    const VectorR<N, E>& n0,
    const VectorR<N, E>& n1,
    const VectorR<N, E>& n2);

  /**
   * Constructs a copy of @a t.
   * The type of the triangle can be different from that of the
   * result triangle as long as the types are assignment compatible.
   */
  TriangleR(
    const TriangleR<N, E>& t);

  /**
   * Assigns this TriangleR from another TriangleR @a t with assignment
   * compatible elements.
   */
  TriangleR<N, E>& operator=(const TriangleR<N, E>& t);

  /**
   * Returns a constant reference to the p0 vector of the triangle.
   */
  const VectorR<N, E>& p0() const
  { return m_p0; }

  /**
   * Returns a constant reference to the p1 vector of the triangle.
   */
  const VectorR<N, E>& p1() const
  { return m_p1; }

  /**
   * Returns a constant reference to the p2 vector of the triangle.
   */
  const VectorR<N, E>& p2() const
  { return m_p2; }

  /**
   * Returns a constant reference to the n0 vector of the triangle.
   */
  const VectorR<N, E>& n0() const
  { return m_n0; }

  /**
   * Returns a constant reference to the n1 vector of the triangle.
   */
  const VectorR<N, E>& n1() const
  { return m_n1; }

  /**
   * Returns a constant reference to the n2 vector of the triangle.
   */
  const VectorR<N, E>& n2() const
  { return m_n2; }

  /**
   * Sets the p0 vector of the triangle.
   */
  void setP0(
    const VectorR<N, E>& p0)
  { m_p0 = p0; }

  /**
   * Sets the p1 vector of the triangle.
   */
  void setP1(
    const VectorR<N, E>& p1)
  { m_p1 = p1; }

  /**
   * Sets the p2 vector of the triangle.
   */
  void setP2(
    const VectorR<N, E>& p2)
  { m_p2 = p2; }

  /**
   * Sets the n0 vector of the triangle.
   */
  void setN0(
    const VectorR<N, E>& n0)
  { m_n0 = n0; }

  /**
   * Sets the n1 vector of the triangle.
   */
  void setN1(
    const VectorR<N, E>& n1)
  { m_n1 = n1; }

  /**
   * Sets the n2 vector of the triangle.
   */
  void setN2(
    const VectorR<N, E>& n2)
  { m_n2 = n2; }

  /**
   * Sets the points and normal of the triangle to the input vectors.
   */
  void set(
    const VectorR<N, E>& p0,
    const VectorR<N, E>& p1,
    const VectorR<N, E>& p2,
    const VectorR<N, E>& n0,
    const VectorR<N, E>& n1,
    const VectorR<N, E>& n2)
  {
    m_p0 = p0;
    m_p1 = p1;
    m_p2 = p2;
    m_n0 = n0;
    m_n1 = n1;
    m_n2 = n2;
  }

  /**
   * Computes the outward pointing normal of the triangle based on the normalized
   * cross product (p2 - p0) x (p1 - p0). This method assumes that the triangle
   * follows a counter-clockwise winding rule.
   */
  void computeNormal();

  /**
   * Computes the area of the triangle based on the triangle points
   */
  double computeArea() const;

  /**
   * Returns true if this triangle equals input triangle @a t.
   */
  bool operator==(
    const TriangleR& t) const;

  /**
   * Returns true if this triangle is not equal to input triangle @a t.
   */
  bool operator!=(
    const TriangleR& t) const
  { return !(*this == t); }

  //---------------------------------------------------------------------------
  // Friend functions
  //---------------------------------------------------------------------------
  /**
   * The method gets the next TriangleR from the input stream and
   * decodes it from little endian format.
   */
  friend void getTriangleR(
    std::istream& inputStream,
    TriangleR<N, E>& t)
  {
    getVectorR(inputStream, t.m_p0);
    getVectorR(inputStream, t.m_p1);
    getVectorR(inputStream, t.m_p2);
    getVectorR(inputStream, t.m_n0);
    getVectorR(inputStream, t.m_n1);
    getVectorR(inputStream, t.m_n2);
  }

  /**
   * The method encodes the input TriangleR into little endian format
   * and writes it to the output stream.
   */
  friend void putTriangleR(
    const TriangleR<N, E>& t,
    std::ostream& outputStream)
  {
    putVectorR(t.m_p0, outputStream);
    putVectorR(t.m_p1, outputStream);
    putVectorR(t.m_p2, outputStream);
    putVectorR(t.m_n0, outputStream);
    putVectorR(t.m_n1, outputStream);
    putVectorR(t.m_n2, outputStream);
  }

private:
  // member variables
  VectorR<N, E> m_p0;
  VectorR<N, E> m_p1;
  VectorR<N, E> m_p2;
  VectorR<N, E> m_n0;
  VectorR<N, E> m_n1;
  VectorR<N, E> m_n2;
};

//---------------------------------------------------------------------------
// Related functions
//---------------------------------------------------------------------------
/**
 * Reports the value of the triangle @a t in text format to
 * the output stream @a out
 *
 * @relates TriangleR
 */
template<unsigned int N, typename E>
std::ostream& operator<<(
  std::ostream& out,
  const TriangleR<N, E> & t);

/**
 * Returns @a true if the points of the two triangles
 * are all within @a epsilon.
 *
 * @relates TriangleR
 */
template<unsigned int N, typename E>
bool almostEqual(
  const TriangleR<N, E>& t0,
  const TriangleR<N, E>& t1,
  double epsilon = pdp::PDP_EPSILON);

// Short type names for common triangle specializations.
typedef TriangleR<2, int> TriangleR2i;
typedef TriangleR<2, unsigned int> TriangleR2ui;
typedef TriangleR<2, float> TriangleR2f;
typedef TriangleR<2, double> TriangleR2d;
typedef TriangleR<3, int> TriangleR3i;
typedef TriangleR<3, unsigned int> TriangleR3ui;
typedef TriangleR<3, float> TriangleR3f;
typedef TriangleR<3, double> TriangleR3d;
typedef TriangleR<4, int> TriangleR4i;
typedef TriangleR<4, unsigned int> TriangleR4ui;
typedef TriangleR<4, float> TriangleR4f;
typedef TriangleR<4, double> TriangleR4d;

} // namespace pdp

#include <elements/TriangleR.t>


