/** @file
 * 
 * Crazyflie manager
 * 
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <vector>
#include <string>
#include <boost/signals2.hpp>
#include <plib/js.h>
#include <models/Robot.h>

class Crazyradio;
class Crazyflie;

namespace pdp {

/**
 * Crazyflie manager object
 *
 * @ingroup models
 */
class PDPMODELS_EXPORT CrazyflieManager: public Robot
{
public:
  enum HoverState
  {
    HoverStateIdle = 0,
    HoverStateLaunching,
    HoverStateHovering,
    HoverStateLanding
  };

  /**
   * Constructs a CrazyflieManager
   */
  explicit CrazyflieManager(
    std::function<void()>* commandResponseFunction,
    std::function<void(std::string)>* stateUpdateFunction,
    std::function<void(const unsigned char*, int, int)>* frameDisplayFunction,
    std::function<void()>* commandsTimeoutFunction,
    std::function<void()>* stateTimeoutFunction,
    std::function<void()>* videoTimeoutFunction,
    float thresholdHeight = 80.0f,
    float launchHeight = 80.0f)
    : Robot(commandResponseFunction, stateUpdateFunction, frameDisplayFunction,
        commandsTimeoutFunction, stateTimeoutFunction, videoTimeoutFunction),
      m_drone(nullptr),
      m_batteryLogBlock(nullptr),
      m_sensorLogBlock(nullptr),
      m_droneConnected(false),
      m_pollingIntervalInSeconds(0.0f),
      m_thresholdHeight(thresholdHeight),
      m_launchHeight(launchHeight),
      m_hoverState(HoverStateIdle)
  {
    m_onBatteryLogFunction = std::bind(&CrazyflieManager::onBatteryLogData, this,
      std::placeholders::_1, std::placeholders::_2);
    m_onSensorLogFunction = std::bind(&CrazyflieManager::onSensorLogData, this,
      std::placeholders::_1, std::placeholders::_2);
  }

  /**
   * Destructs this CrazyflieManager.
   */
  ~CrazyflieManager()
  { disconnect(); }

  virtual bool connect(
    unsigned int pollingIntervalInMilliSeconds) override;

  virtual std::string log() override { return listLogVariables(); }

  virtual std::string parameters() override { return listParameters(); }

  virtual void launch() override;

  virtual void setSlow() override;

  virtual void land() override;

  virtual void stop() override;

  virtual void hover() override;

  virtual void setVelocities(
    float leftRight,
    float backwardForward,
    float downUp,
    float ccwYawCcYaw) override;

  virtual void disconnect() override;

private:
  virtual void ping() override;

  struct batteryLog // Max size 24 bytes - size 6
  {
    float pm_vbat; // 4
    unsigned char pm_batteryLevel; // 1
  } __attribute__((packed));

  struct sensorLog // Max size 24 bytes - size 24
  {
    float stabilizer_roll; // 4
    float stabilizer_pitch; // 4
    float stabilizer_yaw; // 4
    unsigned short range_front; // 2
    unsigned short range_back; // 2
    unsigned short range_up; // 2
    unsigned short range_left; // 2
    unsigned short range_right; // 2
    unsigned short range_zrange; // 2
  } __attribute__((packed));

  bool droneConnected() const
  { return m_drone != nullptr && m_droneConnected; }

  bool startDrone(
    int millisecondsPerPing,
    std::string& radioPowerInfo,
    int& radioChannel,
    std::string& radioDatarateInfo);

  std::string listLogVariables();

  std::string listParameters();

  HoverState hoverState() const
  { return m_hoverState; }

  void navigationControls();

  void stopDrone();

  void onBatteryLogData(
    uint32_t /*time_in_ms*/,
    struct batteryLog* data);

  void onSensorLogData(
    uint32_t /*time_in_ms*/,
    struct sensorLog* data);

private:
  struct BatteryState
  {
    BatteryState()
      : voltage(0.0f),
        level(0.0f)
    {}

    void reset()
    {
      voltage = 0.0f;
      level = 0.0f;
    }

    float voltage;
    float level;
  };

  struct SensorState
  {
    SensorState()
      : roll(0.0f),
        pitch(0.0f),
        yaw(0.0f),
        front(0.0f),
        back(0.0f),
        up(0.0f),
        left(0.0f),
        right(0.0f),
        zrange(0.0f) {}

    void reset()
    {
      roll = 0.0f;
      pitch = 0.0f;
      yaw = 0.0f;
      front = 0.0f;
      back = 0.0f;
      up = 0.0f;
      left = 0.0f;
      right = 0.0f;
      zrange = 0.0f;
    }

    float roll;
    float pitch;
    float yaw;
    float front;
    float back;
    float up;
    float left;
    float right;
    float zrange;
  };

  struct HoverPointTarget
  {
    HoverPointTarget()
      : velocityX(0.0f),
        velocityY(0.0f),
        velocityZ(0.0f),
        yawRate(0.0f),
        height(0.0f) {}

    void reset()
    {
      velocityX = 0.0f;
      velocityY = 0.0f;
      velocityZ = 0.0f;
      yawRate = 0.0f;
      height = 0.0f;
    }

    float velocityX;
    float velocityY;
    float velocityZ;
    float yawRate;
    float height;
  };

  Crazyflie* m_drone;
  LogBlock<struct batteryLog>* m_batteryLogBlock;
  std::function<void(uint32_t, struct batteryLog*)> m_onBatteryLogFunction;
  BatteryState m_batteryState;
  LogBlock<struct sensorLog>* m_sensorLogBlock;
  std::function<void(uint32_t, struct sensorLog*)> m_onSensorLogFunction;
  SensorState m_sensorState;
  bool m_droneConnected;
  float m_pollingIntervalInSeconds;
  float m_thresholdHeight;
  float m_launchHeight;
  HoverState m_hoverState;
  HoverPointTarget m_hoverPointTarget;
};

} // namespace pdp



