/** @file
 *
 * Tello state socket
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <boost/asio.hpp>
#include <utilities/UdpListeningSocket.h>
#include <models/export.h>

using boost::asio::ip::udp;

namespace pdp {

class PDPMODELS_EXPORT TelloStateSocket : public UdpListeningSocket
{
public:
  struct FlightDataHeader {
    std::chrono::system_clock::time_point stamp;
  };

  struct FlightData {
    FlightDataHeader header;
    std::string raw;
    uint8_t sdk;

    // SDK 1.3+ fields
    int pitch;  // pitch in degrees
    int roll;   // roll in degrees
    int yaw;    // yaw in degrees relative to startup orientation
    int vgx;    // x velocity in m/s, +forward
    int vgy;    // y velocity in m/s, +right
    int vgz;    // z velocity in m/s, +down
    int templ;  // lowest temp in C
    int temph;  // highest temp in C
    int tof;    // time of flight distance in cm
    int h;      // height in cm
    int bat;    // battery %
    float baro; // barometric altitude in m
    int time;   // time used by motor in seconds

    // Acceleration includes gravity
    float agx;  // x acceleration in mm/s^2, +forward
    float agy;  // y acceleration in mm/s^2, +right
    float agz;  // z acceleration in mm/s^2, +down

    // SDK 2.0+ fields
    int mid;    // mission pad id, or -1 if not found
    int x;      // x coord relative to mission pad in cm, or 0 if not found
    int y;      // y coord relative to mission pad in cm, or 0 if not found
    int z;      // z coord relative to mission pad in cm, or 0 if not found
  };

  TelloStateSocket(
    unsigned short dataPort,
    std::function<void(std::string)>* stateUpdateFunction);

  virtual ~TelloStateSocket()
  { stopListening(); }

private:
  void processPacket(
    size_t r) override;

private:
  uint8_t m_sdk;  // Tello SDK version
  std::function<void(std::string)>* m_stateUpdateFunction;
};

} // namespace pdp
