/** @file
 *
 * Objects and functions for Cuda texture memory for 3D volume renderering and computation
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <cuda_runtime.h>
#include <elements/MatrixR4x4f.h>
#include <nvidiaTools/export.h>

namespace pdp {

/**
 * @brief CUDA 3d Volume texture
 *
 * Class for implementing a 3d Volume texture using GPU resources and CUDA functions
 *
 * @ingroup neuralNets
 */
class PDPNVIDIATOOLS_EXPORT VolumeCuda3d
{
public:
  /**
   * Constructs VolumeCuda3d
   *
   * @param redVolume pointer to the array of red values
   * @param greenVolume pointer to the array of red values
   * @param blueVolume pointer to the array of red values
   * @param alphaVolume pointer to the array of red values
   * @param width The number of elements in the volume width
   * @param height The number of elements in the volume height
   * @param depth The number of elements in the volume depth
   * @param glPbo OpenGL pixel buffer object to bind the texture to for rendering
   *
   * @note the input element pointers must be pointers to unsigned char arrays of length 'volumeSize'
   */
  VolumeCuda3d(
    void* redVolume,
    void* greenVolume,
    void* blueVolume,
    void* alphaVolume,
    int width,
    int height,
    int depth,
    unsigned int glPbo);

  /**
   * Destructs a VolumeCuda3d.
   */
  ~VolumeCuda3d();

  /**
   * Renders the 2D volume slice defined by positionRatio, channelType, and sliceOrientation. This
   * function operates on the cuda 3D volume and is executed on the GPU via a kernal function
   *
   * @param width Width if the 2d image to display
   * @param height Height if the 2d image to display
   * @param positionRatio Position of the slice in the orthogonal direction epressed from 0 to 1
   * @param channel Which Channel to display (red, green, blue, or composite rgb)
   * @param sliceOrientation The Slice orientation (Axial, Sagittal, Coronal)
   * @param showStructures Boolean indicating whether or not to display the structures (alpha channel)
   * @param structureType Type of structure to display
   * @param structureAlpha The alpha blend for structure display
   */
  void renderCudaSliceGPU(
    unsigned int width,
    unsigned int height,
    float positionRatio,
    int channel,
    int sliceOrientation,
    bool showStructures,
    int structureType,
    float structureAlpha);

  /**
   * Renders the 3D MIP volume defined by channelType, density, and brightness. This
   * function operates on the cuda 3D volume and is executed on the GPU via a kernal function
   *
   * @param width Width if the 2d image to display
   * @param height Height if the 2d image to display
   * @param voxelSizeRatioX Ratio of the 3d Voxel size in the x direction relative to the z dimension
   * @param voxelSizeRatioY Ratio of the 3d Voxel size in the y direction relative to the z dimension
   * @param channel Which Channel to display (red, green, blue, or composite rgb)
   * @param structureType Type of structure to display
   * @param structureAlpha The alpha blend for structure display
   * @param showStructures Boolean indicating whether or not to display the structures (alpha channel)
   * @param density Density of the MIP display
   * @param brightness Brightness of the MIP display
   */
  void renderCudaMipGPU(
    int mipSize,
    const MatrixR4x4f& modelViewMatrix,
    float voxelSizeRatioXtoZ,
    float voxelSizeRatioYtoZ,
    int channel,
    int structureType,
    float structureAlpha,
    bool showStructures,
    float density,
    float brightness);

  bool setupImage2dReader(
    int channels,
    int width,
    int height,
    const VectorR3i& minBounds,
    const VectorR3i& maxBounds,
    int batchSize,
    int& numberOfFilledBatches,
    int& numberOfLeftOverImages);

private:
  // Disable copying
  VolumeCuda3d& operator=(const VolumeCuda3d&) = delete;
  VolumeCuda3d(const VolumeCuda3d&) = delete;

  int iDivUp(
    int a,
    int b);

private:
  // 3D volume textures
  cudaArray* m_gpuRedVolume;
  cudaArray* m_gpuGreenVolume;
  cudaArray* m_gpuBlueVolume;
  cudaArray* m_gpuAlphaVolume;
  VectorR3i m_volumeSizeIDX;
  struct cudaGraphicsResource* m_cudaPboResource;
  int m_batchReadNumberOfImages;
  int m_batchReadImage2dChannels;
  int m_batchReadImage2dWidth;
  int m_batchReadImage2dHeight;
  int m_batchReadImage2dBatchSize;
  VectorR3i m_batchReadPosition;
  int m_batchReadNumberOfImagesRead;
};

} // namespace pdp

