/** @file
 *
 * Interface for cuda / cudnn neural networks
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <stdio.h>
#include <string.h>
#include <iostream>
#include <sstream>
#include <cuda_runtime.h>
#include <cublas_v2.h>
#include <cudnn.h>
#include <curand.h>
#include <curand_kernel.h>
#include <nvidiaTools/export.h>

// Block width for CUDA kernels
const int BW = 128;
const int TILE_DIM = 32;
const int BLOCK_ROWS = 8;

//////////////////////////////////////////////////////////////////////////////
// Error handling
// Adapted from the CUDNN classification code
// sample: https://developer.nvidia.com/cuDNN
#define FatalError(s) do {                                           \
  std::stringstream _where, _message;                                \
  _where << __FILE__ << ':' << __LINE__;                             \
  _message << std::string(s) + "\n" << __FILE__ << ':' << __LINE__;  \
  std::cerr << _message.str() << "\nAborting...\n";                  \
  cudaDeviceReset();                                                 \
  exit(1);                                                           \
} while(0)

#define checkCUDNN(status) do {                                      \
  std::stringstream _error;                                          \
  if (status != CUDNN_STATUS_SUCCESS) {                              \
    _error << "CUDNN error: " << cudnnGetErrorString(status);        \
    FatalError(_error.str());                                        \
  }                                                                  \
} while(0)

#define checkCublasErrors(status) do {                               \
  std::stringstream _error;                                          \
  if (status != 0) {                                                 \
    _error << "Cublas error: " << status;                            \
    FatalError(_error.str());                                        \
  }                                                                  \
} while(0)

#define checkCudaErrors(status) do {                                 \
  std::stringstream _error;                                          \
  if (status != 0) {                                                 \
    _error << "Cuda error: " << cudaGetErrorString(status);          \
    FatalError(_error.str());                                        \
  }                                                                  \
} while(0)


#define CUDA_CALL(function, ...)  {                                  \
  cudaError_t status = function(__VA_ARGS__);                        \
  anyCheck(status == cudaSuccess, cudaGetErrorString(status),        \
    #function, __FILE__, __LINE__);                                  \
}

// This will output the proper error string when calling cudaGetLastError
#define getLastCudaError(msg) __getLastCudaError (msg, __FILE__, __LINE__)
inline void __getLastCudaError(
  const char *errorMessage,
  const char *file,
  const int line)
{
  cudaError_t err = cudaGetLastError();
  if (cudaSuccess != err)
  {
    fprintf(stderr, "%s(%i) : getLastCudaError() CUDA error : %s : (%d) %s.\n",
            file, line, errorMessage, (int)err, cudaGetErrorString(err));
    cudaDeviceReset();
    exit(EXIT_FAILURE);
  }
}

/**
 * Computes ceil(x / y) for integral nonnegative values.
 */
inline unsigned int roundUp(
  unsigned int nominator,
  unsigned int denominator)
{ return (nominator + denominator - 1) / denominator; }

/**
 * Frees memory space in CPU memory
 */
inline void freeCpuMemoryBlock(
  float*& cpuMemoryPointer)
{
  if (cpuMemoryPointer != nullptr)
  {
    checkCudaErrors(cudaFreeHost(cpuMemoryPointer));
    cpuMemoryPointer = nullptr;
  }
}

inline void freeCpuMemoryBlock(
  int*& cpuMemoryPointer)
{
  if (cpuMemoryPointer != nullptr)
  {
    checkCudaErrors(cudaFreeHost(cpuMemoryPointer));
    cpuMemoryPointer = nullptr;
  }
}

/**
 * Frees memory space in GPU memory
 */
inline void freeGpuMemoryBlock(
  float*& gpuMemoryPointer)
{
  if (gpuMemoryPointer != nullptr)
  {
    checkCudaErrors(cudaFree(gpuMemoryPointer));
    gpuMemoryPointer = nullptr;
  }
}

inline void freeGpuMemoryBlock(
  int*& gpuMemoryPointer)
{
  if (gpuMemoryPointer != nullptr)
  {
    checkCudaErrors(cudaFree(gpuMemoryPointer));
    gpuMemoryPointer = nullptr;
  }
}

inline void freeGpuMemoryBlock(
  curandState*& gpuMemoryPointer)
{
  if (gpuMemoryPointer != nullptr)
  {
    checkCudaErrors(cudaFree(gpuMemoryPointer));
    gpuMemoryPointer = nullptr;
  }
}

inline void anyCheck(
  bool is_ok,
  const char *description,
  const char *function,
  const char *file, int line)
{
  if (!is_ok)
  {
    std::cout << "Error: " << description << " in " << function << " at " << file << ":" << line << std::endl;
    exit(EXIT_FAILURE);
  }
}

namespace pdp {

PDPNVIDIATOOLS_EXPORT int reportCudaResources(
  char* cudaResourceInfo);

PDPNVIDIATOOLS_EXPORT int reportGpuDeviceInfo(
  int deviceIdx,
  char* gpuDeviceProperties,
  char* gpuDeviceMemory);

PDPNVIDIATOOLS_EXPORT int getHighestGpuDeviceId();

PDPNVIDIATOOLS_EXPORT int getLowestGpuDeviceId();

PDPNVIDIATOOLS_EXPORT int getGpuWithMostFreeMemoryDeviceId();

inline void resetGpuResources()
{ cudaDeviceReset(); }

PDPNVIDIATOOLS_EXPORT char* cudaResources();

PDPNVIDIATOOLS_EXPORT char* gpuDeviceProperties(
  int deviceIdx);

PDPNVIDIATOOLS_EXPORT char* gpuDeviceMemory(
  int deviceIdx);

} // namespace pdp

