/** @file
 *
 * Agent Class used to implement render / controller model objects into 2D views
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <imaging/VolumeR3.h>
#include <render/Agent.h>
#include <render/View.h>
#include <render/MouseEvent.h>
#include <render/export.h>

namespace pdp {

/**
 * A Agent2d draws a model object in views with orthogonal "flat" projections.
 *
 * @ingroup render
 */
class PDPRENDER_EXPORT Agent2d : public virtual Agent
{
public:
  /**
   * Destructs a Agent2d.
   */
  virtual ~Agent2d();

  /**
   * Draws whatever this agent represents into the specified view.
   * Unless overridden, this function draws nothing.
   */
  virtual void render(
    VolumeR3::RenderOrientation,
    View&) {}

  /**
   * Handles a MouseEvent that occurred in @a view. If this agent accepts
   * the event, it should return true. Otherwise, the event will be sent on
   * to the next agent. Unless overridden, this function returns false,
   * thus rejecting all events.
   *
   * @param event - describes the mouse event that occurred
   * @see MouseEvent
   */
  virtual bool mouseEvent(
    const MouseEvent&,
    View&,
    std::function<void()>)
  { return false; }

  /**
   * Sets the current state of the agent as either enabled or disabled.
   *
   * @note This function is defined as "virtual" so that it can be overriden
   * by derived classes (if specialized code needs to execute during enabling/disabling).
   */
  virtual void setEnabled(
    bool enabled)
  { m_enabled = enabled; }

  /**
   * Returns 'true' if the agent is enabled.
   */
  bool enabled()
  { return m_enabled; }

protected:
  /**
   * Constructs a Agent2d.
   */
  Agent2d()
    : m_enabled(false)
  {}

private:
  bool m_enabled;
};

} // namespace pdp

