/** @file
 *
 * Draw primitives based on OpenGL primitive shaders
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <elements/Color.h>
#include <elements/MatrixR4x4f.h>
#include <models/ModelR3f.h>
#include <render/openglUtils.h>
#include <render/export.h>

namespace pdp
{
  class GLShader;

/**
 * This class supports drawing any set of vertexes in any GL drawing mode.
 * It is intended to replace the user of glBegin/glEnd.  All the vertex lists
 * are drawn using the same model and projection matrix.  Color/opacity varies with
 * the individual vertices's.  After the vertex list has been
 * added, drawing is performed by the draw function. If the geometry doesn't change
 * between frames, draw() may be called again to redraw. This class does not support the lighting models
 * and is not inteded fore 3D surfaces in lighting mode
 *
 * @ingroup render
 */
class PDPRENDER_EXPORT DrawPrimitives2d
{
public:
  DrawPrimitives2d();

  ~DrawPrimitives2d()
  { clearObjectLists(); }

  /**
   * Returns the current default color.
   */
#ifndef _WINDOWS
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wdeprecated-copy"
#endif
  pdp::Color defaultColor() const
  { return m_color; }
#ifndef _WINDOWS
#pragma GCC diagnostic pop
#endif

  /**
   * Determines if the primitives are rendered using the alpha values from the input
   * color values or set to fully opaque.
   */
  void setDefaultColor(
    pdp::Color color,
    bool useDefaultAlpha)
  { m_color = color; m_useDefaultAlpha = useDefaultAlpha; }

  /**
   * Determines if the primitives are rendered using the alpha values from the input
   * color values or set to fully opaque.
   */
  void setDefaultColor(
    const float* color,
    bool useDefaultAlpha)
  { m_color = pdp::Color(static_cast<unsigned char>(color[0] * 255.0f),
      static_cast<unsigned char>(color[1] * 255.0f),
      static_cast<unsigned char>(color[2] * 255.0f),
      static_cast<unsigned char>(color[3] * 255.0f));
    m_useDefaultAlpha = useDefaultAlpha;
  }

  /**
   * Sets the material properties for the lighting model.
   */
  void setMaterial(
    const Material& material)
  { m_material = material; }

  /**
   * Add triangles to the objects lists which uses the default color. The size of the input
   * vector must be multiple of 3 meaning that there are three vertices per triangle
   */
  void addTriangles(
    const std::vector<VectorR3f>& vertices);

  /**
   * Add triangles to the objects lists with colors assigned to each vertex. The size of the input
   * vector must be multiple of 3 meaning that there are three vertices per triangle
   */
  void addTriangles(
    const std::vector<VectorR3f>& vertices,
    pdp::Color color);

  /**
   * Add triangles to the objects lists with colors assigned to each vertex. The size of the input
   * vector must be multiple of 3 meaning that there are three vertices per triangle
   */
  void addTriangles(
    const std::vector<VectorR2f>& vertices,
    pdp::Color color);

  /**
   * Add triangles to the objects lists with colors assigned to each vertex. The size of the input
   * vector must be multiple of 3 meaning that there are three vertices per triangle
   */
  void addTriangles(
    const std::vector<VectorR3f>& vertices,
    const std::vector<pdp::Color>& colors);

  /**
   * Add triangles to the objects lists with a
   */
  void addTriangles(
    const std::vector<VectorR3f>& vertices,
    const std::vector<VectorR3f>& normals);

  /**
   * Adds a triangle strip to the objects lists which uses the default color.
   */
  void addTriangleStrip(
    const std::vector<VectorR3f>& vertices);

  /**
   * Adds a triangle strip to the objects lists with a color assigned to every vertex
   */
  void addTriangleStrip(
    const std::vector<VectorR3f>& vertices,
    pdp::Color color,
    bool stipple);

  /**
   * Adds a triangle strip to the objects lists with a color assigned to every vertex
   */
  void addTriangleStrip(
    const std::vector<VectorR2f>& vertices,
    pdp::Color color,
    bool stipple);

  /**
   * Adds a triangle strip to the objects lists with a color assigned to every vertex
   */
  void addTriangleStrip(
    const std::vector<VectorR3f>& vertices,
    const std::vector<pdp::Color>& colors,
    bool stipple);

  /**
   * Adds a triangle strip to the objects lists which uses the default color.
   */
  void addTriangleStrip(
    const std::vector<VectorR2f>& vertices);

  /**
   * Adds a triangle strip to the objects lists with a color assigned to every vertex
   */
  void addTriangleStrip(
    const std::vector<VectorR2f>& vertices,
    const std::vector<pdp::Color>& colors,
    bool stipple);

  /**
   * Adds a triangle fan to the objects lists which uses the default color.
   */
  void addTriangleFan(
    const std::vector<VectorR3f>& vertices);

  /**
   * Adds a triangle fan to the objects lists with a color assigned to every vertex
   */
  void addTriangleFan(
    const std::vector<VectorR3f>& vertices,
    pdp::Color color);

  /**
   * Adds a triangle fan to the objects lists with a color assigned to every vertex
   */
  void addTriangleFan(
    const std::vector<VectorR2f>& vertices,
    pdp::Color color);

  /**
   * Adds a triangle fan to the objects lists with a color assigned to every vertex
   */
  void addTriangleFan(
    const std::vector<VectorR3f>& vertices,
    const std::vector<pdp::Color>& colors);

  /**
   * Adds a triangle fan to the objects lists which uses the default color.
   */
  void addTriangleFan(
    const std::vector<VectorR2f>& vertices);

  /**
   * Adds a triangle fan to the objects lists with a color assigned to every vertex
   */
  void addTriangleFan(
    const std::vector<VectorR2f>& vertices,
    const std::vector<pdp::Color>& colors);

  /**
   * Add points to the objects lists which uses the default color.
   */
  void addPoints(
    const std::vector<VectorR3f>& vertices,
    float size);

  /**
   * Add points to the objects lists with colors assigned to each vertex.
   */
  void addPoints(
    const std::vector<VectorR3f>& vertices,
    pdp::Color color,
    float size);

  /**
   * Add points to the objects lists with colors assigned to each vertex.
   */
  void addPoints(
    const std::vector<VectorR2f>& vertices,
    pdp::Color color,
    float size);

  /**
   * Add points to the objects lists with colors assigned to each vertex.
   */
  void addPoints(
    const std::vector<VectorR3f>& vertices,
    const std::vector<pdp::Color>& colors,
    float size);

  /**
   * Add lines to the objects lists without colors assigned to each vertex.
   */
  void addLines(
    const std::vector<VectorR2f>& vertices);

  /**
   * Add lines to the objects lists without colors assigned to each vertex.
   */
  void addLines(
    const std::vector<VectorR3f>& vertices);

  /**
   * Add lines to the objects lists with colors assigned to each vertex.
   */
  void addLines(
    const std::vector<VectorR2f>& vertices,
    pdp::Color color,
    bool stipple);

  /**
   * Add lines to the objects lists with colors assigned to each vertex.
   */
  void addLines(
    const std::vector<VectorR3f>& vertices,
    pdp::Color color,
    bool stipple);

  /**
   * Add lines to the objects lists with colors assigned to each vertex.
   */
  void addLines(
    const std::vector<VectorR3f>& vertices,
    const std::vector<pdp::Color>& colors,
    bool stipple);

  /**
   * Add a line loop to the objects lists with colors assigned to each vertex.
   */
  void addLineLoop(
    const std::vector<VectorR2f>& vertices);

  /**
   * Add a line loop to the objects lists with colors assigned to each vertex.
   */
  void addLineLoop(
    const std::vector<VectorR3f>& vertices);

  /**
   * Add a line loop with width to the objects lists with colors assigned to each vertex.
   */
  void addLineLoop(
    const std::vector<VectorR2f>& vertices,
    pdp::Color color,
    bool stipple);

  /**
   * Add a line loop with width to the objects lists with colors assigned to each vertex.
   */
  void addLineLoop(
    const std::vector<VectorR3f>& vertices,
    pdp::Color color,
    bool stipple);

  /**
   * Add a line loop with width to the objects lists with colors assigned to each vertex.
   */
  void addLineLoop(
    const std::vector<VectorR3f>& vertices,
    const std::vector<pdp::Color>& colors,
    bool stipple);

#ifdef STUFF
  /**
   * Add a line strip with width to the objects lists with colors assigned to each vertex.
   */
  void addLineStrip(
    const std::vector<VectorR2f>& vertices,
    const std::vector<pdp::Color>& colors,
    float width,
    bool stipple);

  /**
   * Add a line strip with width to the objects lists with colors assigned to each vertex.
   */
  void addLineStrip(
    const std::vector<VectorR2f>& vertices,
    pdp::Color color,
    float width,
    bool stipple);

  /**
   * Add a line strip with width to the objects lists with colors assigned to each vertex.
   */
  void addLineStrip(
    const std::vector<VectorR3f>& vertices,
    pdp::Color color,
    float width,
    bool stipple);

  /**
  * Add a line strip with width to the objects lists with colors assigned to each vertex.
  */
  void addLineStrip(
    const std::vector<VectorR3f>& vertices,
    const std::vector<pdp::Color>& colors,
    float width,
    bool stipple);

  /**
   * Adds an arrow to the objects lists which uses the default color.
   */
  void addArrow(
    const VectorR2f& start,
    const VectorR2f& end,
    float lineWidth,
    float arrowHeadBaseLength);

  /**
   * Adds an arrow to the objects lists with colors assigned to each vertex.
   */
  void addArrow(
    const VectorR2f& start,
    const VectorR2f& end,
    float lineWidth,
    float arrowHeadBaseLength,
    pdp::Color color);

  /**
   * Adds a circle outline to the objects lists which uses the default color.
   */
  void addCircleOutline(
    const VectorR2f& center,
    float radius,
    float width,
    unsigned int numPoints);

  /**
   * Adds a circle outline to the objects lists with colors assigned to each vertex.
   */
  void addCircleOutline(
    const VectorR2f& center,
    float radius,
    float width,
    unsigned int numPoints,
    pdp::Color color);

  /**
   * Adds a solid circle to the objects lists which uses the default color..
   */
  void addCircle(
    const VectorR2f& center,
    float radius,
    float startAngle,
    float arcSize,
    unsigned int numPoints);

  /**
   * Adds a solid circle to the objects lists with colors assigned to each vertex.
   */
  void addCircle(
    const VectorR2f& center,
    float radius,
    float startAngle,
    float arcSize,
    unsigned int numPoints,
    pdp::Color color);

  /**
   * Adds a navigation point to the objects lists which uses the default color.
   */
  void addNavigationPoint(
    const VectorR2f& center,
    float radius,
    float width,
    unsigned int numPoints,
    pdp::Color color);

  /**
   * Adds a target point to the objects lists with colors assigned to each vertex.
   */
  void addTargetPoint(
    const VectorR2f& center,
    float radius,
    float width,
    unsigned int numPoints,
    pdp::Color color);

  /**
   * Adds a Rounded Rectangle to the objects lists with colors assigned to each vertex.
   */
  void addRoundedRectangle(
    const VectorR2f& position,
    const VectorR2f& size,
    double cornerRadius,
    pdp::Color color);
#endif

  /**
   * Render the current object lists. This function clears the object lists after
   * the render
   */
  void draw(
    const MatrixR4x4f& modelview,
    const MatrixR4x4f& projection,
    GLShader* primitiveShader,
    GLShader* gradientPrimitiveShader,
    GLShader* pointShader,
    GLShader* stippleShader,
    GLShader* lightingShader);

  /**
   * Clears all the pending rendering objects from the objects list without rendering
   */
  void clear()
  { clearObjectLists(); }

private:
  void addLinePrimitives(
    const std::vector<VectorR3f>& vertices);

  void addLinePrimitives(
    const std::vector<VectorR2f>& vertices);

  void addLinePrimitives(
    const std::vector<VectorR3f>& vertices,
    pdp::Color color,
    bool stipple);

  void addLinePrimitives(
    const std::vector<VectorR2f>& vertices,
    pdp::Color color,
    bool stipple);

  void addLinePrimitives(
    const std::vector<VectorR3f>& vertices,
    const std::vector<pdp::Color>& colors,
    bool stipple);

  void addLinePrimitives(
    const std::vector<VectorR2f>& vertices,
    const std::vector<pdp::Color>& colors,
    bool stipple);

  void addLineStripPrimitive(
    const std::vector<VectorR3f>& vertices);

  void addLineStripPrimitive(
    const std::vector<VectorR3f>& vertices,
    const std::vector<pdp::Color>& colors,
    bool stipple);

  void addLineStripPrimitive(
    const std::vector<VectorR2f>& vertices,
    const std::vector<pdp::Color>& colors,
    bool stipple);

  void addLineLoopPrimitive(
    const std::vector<VectorR3f>& vertices);

  void addLineLoopPrimitive(
    const std::vector<VectorR2f>& vertices);

  void addLineLoopPrimitive(
    const std::vector<VectorR3f>& vertices,
    const std::vector<pdp::Color>& colors,
    bool stipple);

  void addLineLoopPrimitive(
    const std::vector<VectorR2f>& vertices,
    pdp::Color color,
    bool stipple);

  void addLineLoopPrimitive(
    const std::vector<VectorR3f>& vertices,
    pdp::Color color,
    bool stipple);

private:
  struct ObjectsGroup
  {
    std::vector<size_t> offsets;
    std::vector<size_t> sizes;
    std::vector<float> vertices;
  };

private:
  void clearObjectLists();

  void drawArrays(
    std::vector<float>& buffer,
    GLuint glOperation,
    GLint vPosition);

  void drawArrays(
    ObjectsGroup& buffer,
    GLuint glOperation,
    GLint vPosition);

  void drawArraysWithColorArrays(
    std::vector<float>& buffer,
    GLuint glOperation,
    GLint vPosition,
    GLint vColor);

  void drawStippleArraysWithColorArrays(
    std::vector<float>& buffer,
    GLuint glOperation,
    GLint vPosition,
    GLint vColor);

  void drawArraysWithColorArrays(
    ObjectsGroup& buffer,
    GLuint glOperation,
    GLint vPosition,
    GLint vColor);

  void drawStippleArraysWithColorArrays(
    ObjectsGroup& buffer,
    GLuint glOperation,
    GLint vPosition,
    GLint vColor);

  void drawArraysWithColorAndSizeArrays(
    std::vector<float>& buffer,
    GLuint glOperation,
    GLint vPosition,
    GLint vColor,
    GLint vSize);

  void drawArraysWithNormalArrays(
    ObjectsGroup& buffer,
    GLuint glOperation,
    GLint vPosition,
    GLint vNormal);

private:
  pdp::Color m_color;
  bool m_useDefaultAlpha;
  Material m_material;
  std::vector<float> m_linesWithoutColors;
  std::vector<float> m_linesWithColors;
  ObjectsGroup m_lineStripsWithoutColors;
  ObjectsGroup m_lineStripsWithColors;
  ObjectsGroup m_lineLoopsWithoutColors;
  ObjectsGroup m_lineLoopsWithColors;
  std::vector<float> m_stippleLinesWithColors;
  ObjectsGroup m_stippleLineStripsWithColors;
  ObjectsGroup m_stippleLineLoopsWithColors;
  std::vector<float> m_trianglesWithoutColors;
  std::vector<float> m_trianglesWithColors;
  ObjectsGroup m_triangleStripsWithoutColors;
  ObjectsGroup m_triangleStripsWithColors;
  ObjectsGroup m_triangleFansWithoutColors;
  ObjectsGroup m_triangleFansWithColors;
  ObjectsGroup m_stippleTriangleStripsWithColors;
  std::vector<float> m_points;
  ObjectsGroup m_trianglesWithNormals;
};

} //namespace pdp


