/** @file
 * 
 * Application object for OpenGL shaders
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <vector>
#include <elements/MatrixR4x4f.h>
#include <render/openglUtils.h>
#include <render/export.h>

namespace pdp {
class GLShaderObject;

/**
 * The GLShader class is used to create and manage executable OpenGL
 * vertex and fragment shaders. This class contains a list of vertex and fragment
 * programs (GLShaderObject objects) and provides functionality to link them
 * together into a shader which can be executed by the OpenGL pipeline.
 *
 * @ingroup render
 */
class PDPRENDER_EXPORT GLShader
{
public:
  /**
   * Creates a GLShader. The caller can check for a successful allocation of
   * OpenGL resources by checking the current OpenGL error message via the
   * @a getGLMessages function.
   */
   GLShader();

  /**
   * Destructs this GLShader. The caller can check for a successful release of
   * OpenGL resources by checking the current OpenGL error message via the
   * @a getGLErrors function.
   */
  ~GLShader();

  /**
   * Adds a shader object (either vertex or fragment) to the current list of
   * shader objects contained in this shader. The input @a shaderObject must
   * be in a compiled state to be added to the shader. The shader takes ownership
   * of the shaderObject memory and will destroy the shaderObject when appropriate.
   * The calling routine must not make modifications or delete the shaderObject
   * after this call. If the shaderObject is OK to add and the shader is already linked,
   * the shader is un-linked (all shaderObjects detached) before the new shaderObject
   * is added. If the function fails it returns @a false, in which case the calling routine
   * should check the current OpenGL error message via the @a getGLMessages function.
   */
   bool addShaderObject(
    GLShaderObject* shaderObject);

  /**
   * Link the currently loaded programs. If the function fails it
   * returns @a false, in which case the calling routine should check the current
   * OpenGL error message via the @a getGLMessages function. Detailed link errors and
   * information can be obtained by checking the linker log via the @a getLinkerLog method
   */
  bool link();

  /**
   * Returns the current link state
   */
  bool linked() const
  { return m_linked; }

  /**
   * Returns the current shader program handle
   */
  GLuint programObject() const
  { return m_programObject; }

  /**
   * This method copies the current shader compiler log into the the @a glShaderInfo
   * string. This log contains detailed information and errors from the last
   * shader compile operation. If the function fails it returns @a false, in which case the
   * calling routine should check the current OpenGL error message via the
   * @a getGLMessages function.
   */
  bool getLinkerLog(
    std::string& glShaderInfo);

  /**
   * Applys the current program so that it is used in place of the regular rendering pipeline.
   */
  bool begin();

  /**
   * Restores the regular rendering pipeline.
   */
  bool end();

  /**
   * Sets float uniform to the program
   */
  bool setUniformFloat( 
    const char* varname, 
    float value);

  /**   
   * Sets a 3 element floting point vector.
   */
  bool setUniformVector3f(
    const char* varname,
    const VectorR3f& vector);

  /**   
   * Sets a 4 element floting point vector.
   */
  bool setUniformVector4f(
    const char* varname,
    const VectorR4f& vector);

  /**
   * Sets int uniform to the program
   */
  bool setUniformInt( 
    const char* varname, 
    int value);

  /**
   * Sets a bool as a uniform
   */
  bool setUniformBool(
    const char* varname,
    bool value);

  /**
   * Sets a pdp::Color as a uniform
   */
  bool setUniformColor(
    const char* varname, 
    const pdp::Color& color);

  /**   
   * Sets a 4x4 floting point Matrix
   */
  bool setUniformMatrix(
    const char* varname,
    const MatrixR4x4f& matrix);

  /**
   * Get the attrib location. an attribute is a variable within a shader
   * that will be bound with data.
   */
  bool getAttribLocation(
    const char* varname,
    GLint& attrib_idx);
  
private:
  GLchar* m_linkerLog;
  bool m_linked;
  GLuint m_programObject;
  std::vector<GLShaderObject*> m_shaderObjects;
};

} // namespace pdp

