/** @file
 *
 * Base class for OpenGL shader objects
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <vector>
#include <render/openglUtils.h>
#include <render/export.h>

namespace pdp {
/**
 * The GLShaderObject class is used to create and manage the
 * source and object (compiled) code for an OpenGL shader. This is a
 * virtual base class for the two types of shaders: @a vertex shaders
 * and @a fragment shaders.
 *
 * @ingroup render
 */
class PDPRENDER_EXPORT GLShaderObject
{
public:
  /**
   * Creates a GLShaderObject
   */
   GLShaderObject();

  /**
   * Destructs this GLShaderObject. The caller can check for a successful release of
   * OpenGL resources by checking the current OpenGL error message via the
   * @a getGLMessages function
   */
  virtual ~GLShaderObject();

  /**
   * Load the source code for this shader from memory. This method
   * does a deep copy of the input memory block and takes ownership of
   * the new memory allocated. After this call returns, the orginal
   * memory block can be released. If the function fails it
   * returns @a false, in which case the calling routine should check the current
   * OpenGL error message via the @a getGLMessages function
   */
  bool loadFromMemory(
    const char* program);

  /**
   * Compile the currently loaded program. If the function fails it
   * returns @a false, in which case the calling routine should check the current
   * OpenGL error message via the @a getGLMessages function. Detailed compile errors and
   * information can be obtained by checking the compiler log via the @a getCompilerLog method
   */
  bool compile();

  /**
   * Returns the current compile state
   */
  bool compiled() const
  { return m_compiled; }

  /**
   * Returns the current shader object handle
   */
  GLuint shaderObject() const;

  /**
   * This method copies the current shader compiler log into the the @a glShaderInfo
   * string. This log contains detailed information and errors from the last
   * shader compile operation. If the function fails it returns @a false, in which case the
   * calling routine should check the current OpenGL error message via the
   * @a getGLMessages function.
   */
  bool getCompilerLog(
    std::string& glShaderInfo);

  /**
   * Returns the Attribute Location specify by the input attribute name @a attributeName.
   * shader compile operation.
   */
  int getAttributeLocation(
    char* attributeName);

protected:
  GLuint m_shaderObject;

private:
  char* m_shaderSource;
  char* m_compilerLog;
  bool m_compiled;
};

/**
 * The GLVertexShaderObject class is used to create and manage the
 * source and object (compiled) code for an OpenGL vertex shader.
 * This class derives most of its functionality from the @a GLShaderObject class
 *
 * @ingroup render
 */
class PDPRENDER_EXPORT GLVertexShaderObject : public GLShaderObject
{
public:
  /**
   * Creates a GLVertexShaderObject. The caller can check for a successful allocation of
   * OpenGL resources by checking the current OpenGL error message via the
   * @a getGLMessages function
   */
   GLVertexShaderObject();
};

/**
 * The GLFragmentShaderObject class is used to create and manage the
 * source and object (compiled) code for an OpenGL fragment shader.
 * This class derives most of its functionality from the @a GLShaderObject class
 *
 * @ingroup render
 */
class PDPRENDER_EXPORT GLFragmentShaderObject : public GLShaderObject
{
public:
  /**
   * Creates a GLFragmentShaderObject. The caller can check for a successful allocation of
   * OpenGL resources by checking the current OpenGL error message via the
   * @a getGLMessages function
   */
   GLFragmentShaderObject();
};

} // namespace pdp

