/** @file
 *
 * Rendering engine for 2D images. This Class contains
 * all the objects such as OpenGL Context and image container for rendering 2D images.
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <boost/signals2.hpp>
#include <utilities/timeUtils.h>
#include <imaging/ImageR2.h>
#include <imaging/Image2dDetectionsAdaptor.h>
#include <render/RenderContext.h>
#include <render/openglUtils.h>
#include <render/export.h>

using namespace std;
using namespace std::chrono;

namespace pdp {

class View;
class NavigationPointAgent2d;
class ViewLayout;

/**
 * Rendering engine for 2D images. This Class contains
 * all the objects such as OpenGL Context and image container for rendering 2D images.
 * The engine stores all the volumes and generates 2d images using an offscreen OpenGL Context
 *
 * @ingroup render
 */
class PDPRENDER_EXPORT ImageRenderingEngine
{
public:
  /**
   * Constructs a ImageRenderingEngine
   */
  ImageRenderingEngine(
    int width,
    int height,
    bool createJpegImage);

  /**
   * Destructs a ImageRenderingEngine.
   */
  ~ImageRenderingEngine();

  void setEndpoint(
    const std::string& remoteEndpointAddress,
    unsigned short remoteEndpointPort)
  {
    m_remoteEndpointAddress = remoteEndpointAddress;
    m_remoteEndpointPort = remoteEndpointPort;
  }

  /**
   * Returns the current remote endpoint address
   */
  const std::string& remoteEndpointAddress() const
  { return m_remoteEndpointAddress; }

  /**
   * Returns the current remote endpoint port
   */
  unsigned short remoteEndpointPort() const
  { return m_remoteEndpointPort; }

  /**
   * Returns the current engine information
   */
  std::string info() const
  { return m_info; }

  RenderContext* context()
  { return m_context; }

  /**
   * Returns the current context information
   */
  std::string contextInfo() const
  { return m_context->glInfo(); }

  /**
   * Returns the current image information
   */
  std::string imageInfo() const
  { return m_image->info(); }

  /**
   * Returns a reference to the current image
   */
  ImageR2* image()
  { return m_image; }

  /**
   * Loads a image specified by the input full file path and name. The type of the image is determined
   * by the file's extension. PNG ".png" and Jpeg ".jpg" formats are currently supported
   *
   * @param filePathAndName
   * @return 'true' on successful image load
   */
  bool loadImage(
    const std::string& filePathAndName,
    const VectorR2f& resolution,
    bool avgCompress2x2);

  bool loadDetectionsImage(
    const unsigned char* detectionsBuffer,
    const VectorR3i& size,
    const VectorR2f& resolution,
    bool avgCompress2x2,
    const std::vector<Color>& colorLookup,
    bool linearInterp);

  /**
   * Unloads all images and frees their data structures
   */
  void unloadImage();

  void unloadDetectionsImage();

  void attachImageDetections(
    Image2dDetectionsAdaptor::ClassificationType classificationType,
    Image2dDetectionsAdaptor* imageDetections)
  {
    m_classificationType = classificationType;
    m_imageDetections = imageDetections;
  }

  void detachImageDetections()
  { m_imageDetections = nullptr; }

  /**
   * Returns 'true' if rendering is enabled meaning that the context is valid
   */
  bool renderingEnabled() const
  { return m_context->renderingEnabled(); }

  /**
   * Returns 'true' if images are loaded and the imageTexture has been successfully loaded
   */
  bool imageLoaded() const
  { return m_image->imageLoaded(); }

  /**
   * Returns 'true' if the detections image is loaded and the imageTexture has been successfully loaded
   */
  bool detectionsImageLoaded() const
  { return m_detectionsImage->imageLoaded(); }

  /**
   * Renders the current state of the rendering engine and updates the internal output image buffers with
   * the result.
   */
  void render(
    Image2dDetectionsAdaptor::ClassificationType classificationType,
    pdp::ImageR2::StructureChannel structureChannel,
    const std::vector<pdp::Color>& classColors,
    const std::vector<bool>& showClasses,
    const VectorR2f& displaySample);

  void renderDetectionsImage(
    const VectorR2f& displaySample);

  /**
   * Returns the current uncompressed image buffer.
   */
  const unsigned char* uncompressedImage()
  { return m_uncompressedImage; }

  int width() const
  { return m_context->width(); }

  int height() const
  { return m_context->height(); }

  const unsigned char* jpegImage() const
  { return m_jpegImage; }

  unsigned int jpegImageLength() const
  { return m_jpegImageLength; }

  float renderTime() const
  { return (float)m_renderTime.count() * 1.0f; }

  float compressionTime() const
  { return (float)m_compressionTime.count() * 1.0f; }

  /**
   * Interprets mouse press events and passes them along to the views.
   * @a position is specified in window coordinates. @a doubleClick should
   * be true if this press event is the second one in a double click sequence
   * (mousePressEvent()should be be called also for the first one). Returns
   * @a true if the layout accepts the event, else it returns false.
   */
  bool mousePressEvent(
    const VectorR2i& position,
    MouseEvent::ButtonState button,
    MouseEvent::ModifierKeyState modifiers,
    MouseEvent::Interpretation interpretation,
    std::function<void()> updateRender);

  /**
   * Interprets mouse release events and passes them along to the views.
   * @a position is given in window coordinates. Returns @a true
   * if the layout accepts the event, else it returns false.
   */
  bool mouseReleaseEvent(
    const VectorR2i& position,
    MouseEvent::ButtonState button,
    MouseEvent::ModifierKeyState modifiers,
    std::function<void()> updateRender);

  /**
   * Interprets mouse motion events and passes them along to the views.
   * @a position is given in window coordinates. Returns @a true
   * if the layout accepts the event, else it returns false.
   */
  bool mouseMoveEvent(
    const VectorR2i& position,
    std::function<void()> updateRender);

private:
  std::string m_remoteEndpointAddress;
  unsigned short m_remoteEndpointPort;
  RenderContext* m_context;
  View* m_views[1];
  NavigationPointAgent2d* m_navigationAgent;
  ViewLayout* m_viewLayouts[1];
  ImageR2* m_image;
  ImageR2* m_detectionsImage;
  Image2dDetectionsAdaptor::ClassificationType m_classificationType;
  Image2dDetectionsAdaptor* m_imageDetections;
  std::string m_info;
  unsigned char* m_uncompressedImage;
  bool m_createJpegImage;
  unsigned char* m_jpegImage;
  unsigned int m_jpegImageLength;
  std::chrono::duration<double, std::milli> m_renderTime;
  std::chrono::duration<double, std::milli> m_compressionTime;
  std::chrono::high_resolution_clock::time_point m_lastRenderTime;
  double m_renderingInterval;
};

} // namespace pdp

