/** @file
 *
 * Stores a image in OpenGL texture memory for 2D image renderering
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <render/openglUtils.h>
#include <render/export.h>

namespace pdp {

class ImageR2;

/**
 * Stores a image in OpenGL texture memory for 2D image rendering
 *
 * @ingroup render
 */
class PDPRENDER_EXPORT ImageTexture2d
{
public:
  /**
   * Enumerated type used to identify the current state of the texture image
   */
  enum ImageTextureState
  {
    ImageTextureEmpty = 0,
    ImageTextureLoaded
  };

  /**
   * Constructs a ImageTexture2d which stores a 2D image from the input
   * image in OpenGL texture memory. The ImageTexture2d is initially empty
   */
  ImageTexture2d();

  /**
   * Destructs a ImageTexture2d.
   */
  ~ImageTexture2d();

  /**
   * Returns the current state of the texture image
   */
  ImageTextureState state() const
  { return m_state; }

  /**
   * Returns the current requested state of the texture image
   */
  ImageTextureState requestedState() const
  { return m_requestedState; }

  /**
   * Loads a ImageTexture from the input @a image.
   * The resolution is based on the input @a image size. This is not an immediate
   * load but simply saves the load parameters for the next display cycle.
   */
  void loadFromImage(
    ImageTextureState requestedState,
    ImageR2* image,
    bool linearInterp);

#ifdef TODO
  /**
   * Clears the structure overlay.
   */
  void clearOverlay();

  /**
   * Loads the classification stored in the current image into the 2D texture
   * for display.
   */
  void loadClassificationFromImage();

  /**
   * Loads the truth stored in the current image into the 2D texture
   * for display.
   */
  void loadTruthFromImage();
#endif

  /**
   * Returns a pointer to the current image for the image display texture. This
   * pointer can be nullptr and must be checked by the caller
   */
  ImageR2* image()
  { return m_image; }

  /**
   * Unloads the texture and frees up the texture resources
   */
  void unload();

  /**
   * Returns @a true if the image is currently loaded
   */
  bool imageLoaded() const
  { return m_imageLoaded; }

  /**
   * Returns the OpenGL pixel buffer object Id
   */
  GLuint glPbo() const
  { return m_glPbo; }

  /**
   * Returns the OpenGL 2d texture object Id
   */
  GLuint glTexture2d() const
  { return m_glTexture2d; }

  /**
   * Immediate update of the texture. This method should be called at
   * render time to update the texture image.
   */
  void updateTexture();

  /**
   * Immediate update of the texture parameters. This method should be called at
   * render time to update display properties as needed.
   */
  void updateProperties();

private:
  ImageTextureState m_requestedState;
  ImageTextureState m_state;
  ImageR2* m_image;
  bool m_linearInterp;
  bool m_imageLoaded;
  GLuint m_glPbo;  // OpenGL pixel buffer object
  GLuint m_glTexture2d; // OpenGL texture object
  bool m_loadNeedsUpdate;
  bool m_propertiesNeedUpdate;
};

} // namespace pdp

