/** @file
 *
 * Mouse events
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <elements/VectorR.h>
#include <render/export.h>

namespace pdp {

/**
 * Represents events generated by a pointer device.
 *
 * @ingroup render
 */
class PDPRENDER_EXPORT MouseEvent
{
public:
  /**
   * Identifies the supported cursor shapes.
   */
  typedef enum
  {
    PDP_ARROW_CURSOR = 0,
    PDP_WAIT_CURSOR,
    PDP_POINTING_CURSOR,
    PDP_FORBIDDEN_CURSOR,
    PDP_QUESTION_CURSOR,
    PDP_NO_CURSOR,
    PDP_CUT_CURSOR,
    PDP_DRAW_CURSOR,
    PDP_ERASE_CURSOR,
    PDP_PAINT_CURSOR,
    PDP_SHAPE_CURSOR,
    PDP_BRUSH_CURSOR,
    PDP_BRUSH_FILL_CURSOR,
    PDP_SMART_BRUSH_CURSOR,
    PDP_SMART_ERASE_CURSOR
  } CursorType;

  /**
   * Identifies the specific type of mouse event.
   */
  enum Type
  {
    PRESS = 1,   /**< A mouse button was pressed. */
    RELEASE = 2, /**< A mouse button was released. */
    MOVE = 4,    /**< The mouse was moved. */
    ENTER = 8,   /**< The mouse cursor entered the a view or window. */
    LEAVE = 16   /**< The mouse cursor left the a view or window. */
  };

  /**
   * A set of flags describing possible interpretations of a low-level event.
   */
  typedef unsigned int Interpretation;

  /**
   * Enumerated #Interpretation values.
   */
  enum
  {
    PDP_NO_INTERPRETATION = 0,
    PDP_DRAG_BEGIN   = 1, /**< The PRESS event marked the beginning of a drag operation. */
    PDP_DRAGGING     = 2, /**< The MOVE event occurred during dragging. */
    PDP_DRAG_END     = 4, /**< The RELEASE event completed a drag operation. */
    PDP_DOUBLE_CLICK = 8  /**< The PRESS event was the second in a double click. */
  };

  /**
   * Identifies a set of mouse buttons.
   */
  typedef unsigned ButtonState;

  /**
   * Enumerated #ButtonState values.
   */
  enum
  {
    NO_BUTTON     = 0,
    LEFT_BUTTON   = 1, /**< Left mouse button. */
    RIGHT_BUTTON  = 2, /**< Right mouse button. */
    MIDDLE_BUTTON = 4  /**< Middle button (on a three-button mouse). */
  };

  /**
   * A set of modifier keys.
   */
  typedef unsigned ModifierKeyState;

  /**
   * ModifierKeyState symbols.
   */
  enum
  {
    NO_KEY      = 0,
    SHIFT_KEY   = 1, /**< Shift key. */
    CONTROL_KEY = 2, /**< Control key. */
    ALT_KEY     = 4  /**< Alt key. */
  };

  /**
   * Constructs a MouseEvent.
   *
   * @param type - the specific type of mouse event
   * @param interpretation - a set of possible interpretations of this event
   * @param position - the view or window position in which the event occurred
   * @param button - the button that caused a PRESS or RELEASE event
   * @param modifiers - the modifier keys that were pressed just before
   * the event was generated
   */
  MouseEvent(
    Type type,
    Interpretation interpretation,
    const VectorR2f& position,
    ButtonState button = 0,
    ModifierKeyState modifiers = 0);

  /**
   * Returns the type of view event.
   */
  Type type() const;

  /**
   * Returns bits that describes the interpretation of this event.
   */
  Interpretation interpretation() const;

  /**
   * Returns the current mouse position for a PRESS, RELEASE or MOVE
   * view event.
   */
  const VectorR2f& position() const;

  /**
   * Returns the mouse button that caused a PRESS or RELEASE view event.
   */
  ButtonState button() const;

  /**
   * Returns the set of modifier keys that were pressed when a PRESS
   * or RELEASE view event was generated.
   */
  ModifierKeyState modifiers() const;

private:
  Type m_type;
  Interpretation m_interpretation;
  VectorR2f m_position;
  int m_button;
  ModifierKeyState m_modifiers;
};

} // namespace pdp

