/** @file
 *
 * Agent Class used to implement render / controller navigation points into 2d views
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2020
 */
#pragma once

#include <elements/VectorR.h>
#include <elements/Color.h>
#include <render/Agent2d.h>

namespace pdp {

/**
 * Agent used to display a navigation point and to change its position
 * and the positions of the displayed planes in 2D views. The point
 * is moved in the plane of the rendered view, while the viewing planes
 * are moved in the direction of their normal vector by a magnitude of the
 * movement vector along the plane normal.
 *
 * @ingroup render
 */
class PDPRENDER_EXPORT NavigationPointAgent2d : virtual public Agent2d
{
public:
  /**
   * Constructs a NavPointRenderer2d that draws the navigation point in one
   * or several 2D views, and changes the position of the attached NavigationPoint
   * and viewing planes through mouse events.
   */
  NavigationPointAgent2d(
    float selectTolerance = 12.0f,
    Color markerColor = Color(0, 255, 100, 180),
    unsigned int markerSize = 9,
    Color selectColor = Color(255, 0, 0, 180),
    float orthoSensitivity = 0.2f);

  /**
   * Destroys a NavPointRenderer2d
   */
  ~NavigationPointAgent2d();

  /**
   * Attach the navigation point to a volume and initialize it's navigation position
   * to the center of the volume.
   */
  void attachVolume(
    VolumeR3* volume);

  const VectorR3f& pointSCE() const
  { return m_pointSCE; }

  void render(
    VolumeR3::RenderOrientation renderOrientation,
    View& view) override;

  bool mouseEvent(
    const MouseEvent& event,
    View& view,
    std::function<void()> updateRender) override;

private:
  // Navigation objects
  Color m_markerColor;
  unsigned int m_markerSize;
  Color m_selectColor;
  float m_selectTolerance;
  float m_orthoSensitivity;
  bool m_ortho;
  bool m_disableNavigationX;
  bool m_disableNavigationY;
  VolumeR3* m_volume;
  VectorR3f m_pointSCE;
  bool m_moving;
  VectorR2f m_currentVPT;
  View* m_viewOnMouseDown;
};

} // namespace pdp

