/** @file
 *
 * OpenGL offscreen rendering context
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <string>
#include <vector>
#include <array>
#include <stack>
#include <imaging/VolumeR3.h>
#include <imaging/ImageR2.h>
#ifdef PDP_OS_UBUNTU
#include <nvidiaTools/OptixRenderer.h>
#endif
#include <render/openglUtils.h>
#include <render/DrawPrimitives2d.h>
#include <render/VolumeTexture3d.h>
#include <render/ImageTexture2d.h>
#include <render/MouseEvent.h>
#include <render/export.h>

namespace pdp {

class GLShader;
class GLFragmentShaderObject;
class GLVertexShaderObject;
class ModelR3f;

/**
 *
 * @ingroup render
 */
class PDPRENDER_EXPORT RenderContext
{
public:
  /**
   * Enumerated type used to specify a volume shader
   */
  enum RenderMode
  {
    RenderModeVolume = 0,
    RenderModeImage,
    RenderModePrimatives,
    RenderModeRayTrace
  };

  /**
   * Enumerated type used to specify a volume shader
   */
  enum ShaderName
  {
    PrimitiveShader = 0,
    GradientPrimitiveShader,
    Texture2dShader,
    LightingShader,
    PointShader,
    StippleShader,
    ShaderEnumListSize // This must remain the last shader, to keep the array size correct
  };

  RenderContext(
    int width,
    int height,
    RenderMode renderMode);

  ~RenderContext();

  GLint width() const
  { return m_width; }

  GLint height() const
  { return m_height; }

  RenderMode renderMode() const
  { return m_renderMode; }

  bool renderingEnabled() const
  { return m_renderingEnabled; }

  /**
   * Returns the current OpenGL information
   */
  const std::string& glInfo() const
  { return m_glInfo; }

  /**
   * Returns a pointer to the shader with name @a name. The calling rouint
   * should check for a nullptr return pointer before using the shader
   */
  GLShader* shader(
    ShaderName name)
  { return m_shaders[name]; }

  void renderModelParts(
    ModelR3f* model);

  /**
   * Loads the input @a fixedVolume into the context texture memory for volume rendering
   */
  void loadFixedVolumeTexture(
    VolumeR3* fixedVolume,
    //const MapRusToRGBA* fixedLookupTable,
    bool linearInterp);

  /**
   * Unloads the context volume texture
   */
  void unloadVolumeTexture();

  /**
   * Updates the volume texture at render time if needed.
   */
  void updateVolumeTexture();

  /**
   * Updates the volume texture paramters at render time if needed.
   */
  void updateVolumeTextureProperties();

  /**
   * Loads the input @a Image into the context texture memory for image rendering
   */
  void loadImageTexture(
    ImageR2* image,
    bool linearInterp);

  /**
   * Unloads the context image texture
   */
  void unloadImageTexture();

  /**
   * Updates the image texture at render time if needed.
   */
  void updateImageTexture();

  /**
   * Updates the image texture parameters at render time if needed.
   */
  void updateImageTextureProperties();

  /**
   * Loads the input @a detections texture into the context texture memory for image rendering
   */
  void loadDetectionsTexture(
    ImageR2* detectionsImage,
    bool linearInterp);

  /**
   * Unloads the context Detections texture
   */
  void unloadDetectionsTexture();

  /**
   * Updates the Detections texture at render time if needed.
   */
  void updateDetectionsTexture();

  /**
   * Updates the Detections texture parameters at render time if needed.
   */
  void updateDetectionsTextureProperties();

  /**
   * Returns constant reference to the projection matrix.
   */
  const MatrixR4x4f& projectionMatrix() const
  { return m_projection.top().xForm; }

  /**
   * Set the global projection transform to identity
   */
  void setProjectionIdentity()
  { matrixR4x4f_identity(m_projection.top().xForm); }

  /**
   * Set the global projection transform
   */
  void setProjection(
	  MatrixR4x4f& fwdXform)
  { matrixR4x4f_dup(m_projection.top().xForm, fwdXform); }

  /**
   * Scale the global projection transform by the input vector
   */
  void scaleProjection(
    const VectorR3f& scaling)
  {
    MatrixR4x4f scaleMatrix;
    matrixR4x4f_identity(scaleMatrix);
    scaleMatrix[0][0] = scaling[X];
    scaleMatrix[1][1] = scaling[Y];
    scaleMatrix[2][2] = scaling[Z];
    scaleMatrix[3][3] = 1.0f;
    matrixR4x4f_mul(m_projection.top().xForm, scaleMatrix, m_projection.top().xForm);
  }

  /**
   * Scale the global projection transform by the input vector
   */
  void scaleProjection(
    float x,
    float y,
    float z)
  {
    MatrixR4x4f scaleMatrix;
    matrixR4x4f_identity(scaleMatrix);
    scaleMatrix[0][0] = x;
    scaleMatrix[1][1] = y;
    scaleMatrix[2][2] = z;
    scaleMatrix[3][3] = 1.0f;
    matrixR4x4f_mul(m_projection.top().xForm, scaleMatrix, m_projection.top().xForm);
  }

  /**
   * Rotate the global projection transform around the input axis by the input angle
   */
  void rotateProjection(
    float angle,
    const VectorR3f& axis)
  { matrixR4x4f_rotate(m_projection.top().xForm, m_projection.top().xForm, axis[X], axis[Y], axis[Z], (float)(angle * PDP_DEG2RAD)); }

  /**
   * translate the global projection transform by the input vector
   */
  void translateProjection(
    const VectorR3f& translation)
  { matrixR4x4f_translate_in_place(m_projection.top().xForm, translation[X], translation[Y], translation[Z]); }

  /**
   * translate the global projection transform by the input vector
   */
  void translateProjection(
    float x,
    float y,
    float z)
  { matrixR4x4f_translate_in_place(m_projection.top().xForm, x, y, z); }

  /**
   * Concatinate the global projection transform by the input matrix
   */
  void multProjection(
    MatrixR4x4f& xform)
  { matrixR4x4f_mul(m_projection.top().xForm, xform, m_projection.top().xForm); }

  /**
   * Creates a copy of the current projection and pushes it onto the projection stack
   */
  void pushProjection()
  { m_projection.push(m_projection.top()); }

  /**
   * Pops the current projection off od the projection stack
   */
  void popProjection()
  { if (m_projection.size() > 1) m_projection.pop(); }

  /**
   * Returns constant reference to the projection matrix.
   */
  const MatrixR4x4f& modelviewMatrix() const
  { return m_modelview.top().xForm; }

  /**
   * Set the global modelview transform to identity
   */
  void setModelviewIdentity()
  { matrixR4x4f_identity(m_modelview.top().xForm); }

  /**
   * Set the global modelview transform
   */
  void setModelview(
	  MatrixR4x4f& fwdXform)
  { matrixR4x4f_dup(m_modelview.top().xForm, fwdXform); }

  /**
   * Scale the global modelview transform by the input vector
   */
  void scaleModelview(
    const VectorR3f& scaling)
  {
    MatrixR4x4f scaleMatrix;
    matrixR4x4f_identity(scaleMatrix);
    scaleMatrix[0][0] = scaling[X];
    scaleMatrix[1][1] = scaling[Y];
    scaleMatrix[2][2] = scaling[Z];
    scaleMatrix[3][3] = 1.0f;
    matrixR4x4f_mul(m_modelview.top().xForm, scaleMatrix, m_modelview.top().xForm);
  }

  /**
   * Scale the global modelview transform by the input vector
   */
  void scaleModelview(
    float x,
    float y,
    float z)
  {
    MatrixR4x4f scaleMatrix;
    matrixR4x4f_identity(scaleMatrix);
    scaleMatrix[0][0] = x;
    scaleMatrix[1][1] = y;
    scaleMatrix[2][2] = z;
    scaleMatrix[3][3] = 1.0f;
    matrixR4x4f_mul(m_modelview.top().xForm, scaleMatrix, m_modelview.top().xForm);
  }

  /**
   * Rotate the global projection transform around the input axis by the input angle
   */
  void rotateModelview(
    float angle,
    const VectorR3f& axis)
  { matrixR4x4f_rotate(m_modelview.top().xForm, m_modelview.top().xForm, axis[X], axis[Y], axis[Y], (float)(angle * PDP_DEG2RAD)); }

  /**
   * translate the global modelview transform by the input vector
   */
  void translateModelview(
    const VectorR3f& translation)
  { matrixR4x4f_translate_in_place(m_modelview.top().xForm, translation[X], translation[Y], translation[Z]); }

  /**
   * translate the global modelview transform by the input vector
   */
  void translateModelview(
    float x,
    float y,
    float z)
  { matrixR4x4f_translate_in_place(m_modelview.top().xForm, x, y, z); }

  /**
   * Concatinate the global modelview transform by the input matrix
   */
  void multModelview(
    MatrixR4x4f& xform)
  { matrixR4x4f_mul(m_modelview.top().xForm, xform, m_modelview.top().xForm); }

  /**
   * Multiply the global modelview transform by the input matrix
   */
  void multModelviewInPlace(
    MatrixR4x4f& xform)
  { matrixR4x4f_mul_in_place(m_modelview.top().xForm, xform); }

  /**
   * Creates a copy of the current modelview and pushes it onto the modelview stack
   */
  void pushModelview()
  { m_modelview.push(m_modelview.top()); }

  /**
   * Pops the current modelview off od the modelview stack
   */
  void popModelview()
  { if (m_modelview.size() > 1) m_modelview.pop(); }

  void captureCurrentRendering(
    const VectorR2i& position,
    const VectorR2i& size,
    const VectorR2f& contextOffset,
    float zoom,
    float width,
    float height);

  /**
   * Returns a writable reference to the 3d volume texture
   */
  VolumeTexture3d* volumeTexture()
  { return m_volumeTexture; }

  /**
   * Returns a writable reference to the 2d image texture
   */
  ImageTexture2d* imageTexture()
  { return m_imageTexture; }

  /**
   * Returns a writable reference to the 2d detections texture
   */
  ImageTexture2d* detectionsTexture()
  { return m_detectionsTexture; }

  /**
   * Returns a writable reference to the 2d object render for rendering objects composed of 2D primitives
   */
  DrawPrimitives2d& objectRenderer2d()
  { return m_objectRenderer2d; }

#ifdef PDP_OS_UBUNTU
  OptixRenderer* rayTracer() const
  { return m_rayTracer; }
#endif

  /**
   * Draws the objects currently queded in the object renderer object lists and then clears these lists
   */
  void drawObjects2d();

private:
#ifndef PDP_OS_UBUNTU // MacOSX and Windows
#else // UBUNTU
  void setupEGL(
    int width,
    int height);

  void shutdownEGL();
#endif
  void unloadShader(
    ShaderName name);

  void unloadShaders();

  bool setShader(
    ShaderName name,
    std::string& glInfo);

  GLVertexShaderObject* createVertexShader(
    std::string &glInfo,
    const std::string& shaderName,
    const char* vertexProgram);

  GLFragmentShaderObject* createFragShader(
    std::string &glInfo,
    const std::string& shaderName,
    const char* fragmentProgram);

  void updateContext() {}

private:
  struct XForm
  {
    MatrixR4x4f xForm;
  };

  GLint m_width;
  GLint m_height;
  RenderMode m_renderMode;
#if defined(PDP_OS_WINDOWS)
  GLFWwindow* m_window;
#elif defined(PDP_OS_MACOSX)
  GLFWwindow* m_window;
#else // Ubuntu
  EGLConfig m_eglConf;
  EGLSurface m_eglSurface;
  EGLContext m_eglCtx;
  EGLDisplay m_eglDisp;
#endif
  bool m_renderingEnabled;
  GLuint m_buffersIdx;
  GLuint m_vertexArraysIdx;
  GLuint m_texcoordsIdx;
  GLuint m_verticesIdx;
  std::string m_glInfo;
  std::stack<XForm> m_projection;
  std::stack<XForm> m_modelview;
  DrawPrimitives2d m_objectRenderer2d;

  // Rendering objects
  VolumeTexture3d* m_volumeTexture;
  ImageTexture2d* m_imageTexture;
  ImageTexture2d* m_detectionsTexture;
  std::array<GLShader*, ShaderEnumListSize> m_shaders;
#ifdef PDP_OS_UBUNTU
  Camera m_camera;
  OptixModel* m_model;
  OptixRenderer* m_rayTracer;
#endif
};

} // namespace pdp

