/** @file
 * 
 * View Layout Class used to implement view layouts for OpenGL rendering
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <vector>
#include <render/MouseEvent.h>
#include <render/export.h>

namespace pdp {

class View;

/**
 * View Layout Class used to implement view layouts for OpenGL rendering
 *
 * @ingroup render
 */
class PDPRENDER_EXPORT ViewLayout
{
public:
  /**
   * Constructs a new ViewLayout.
   */
  ViewLayout(
    std::string label)
    : m_label(label)
  {}

  /**
   * Destructs this ViewLayout.
   */
  ~ViewLayout()
  {}

  void addView(
    View* view);

  /**
   * Interprets mouse press events and passes them along to the views.
   * @a position is specified in window coordinates. @a doubleClick should
   * be true if this press event is the second one in a double click sequence
   * (mousePressEvent()should be be called also for the first one). Returns
   * @a true if the layout accepts the event, else it returns false.
   */
  bool mousePressEvent(
    const VectorR2i& position,
    MouseEvent::ButtonState button, 
    MouseEvent::ModifierKeyState modifiers,
    MouseEvent::Interpretation interpretation,
    std::function<void()> updateRender);
  
  /**
   * Interprets mouse release events and passes them along to the views.
   * @a position is given in window coordinates. Returns @a true
   * if the layout accepts the event, else it returns false.
   */
  bool mouseReleaseEvent(
    const VectorR2i& position,
    MouseEvent::ButtonState button, 
    MouseEvent::ModifierKeyState modifiers,
    std::function<void()> updateRender);
  
  /**
   * Interprets mouse motion events and passes them along to the views.
   * @a position is given in window coordinates. Returns @a true
   * if the layout accepts the event, else it returns false.
   */
  bool mouseMoveEvent(
    const VectorR2i& position,
    std::function<void()> updateRender);

private:
  const VectorR2f windowToViewportCoordinate(
    const VectorR2i& pos,
    const View* view) const;

private:
  std::string m_label;
  std::vector<View*> m_views;
};

} // namespace pdp

