/** @file
 *
 * Rendering engine for 3D tomographic medical image volumes. This Class contains
 * all the objects such as OpenGL Context and volume container for rendering volumes.
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <boost/signals2.hpp>
#include <utilities/timeUtils.h>
#include <imaging/VolumeR3.h>
#include <render/RenderContext.h>
#include <render/openglUtils.h>
#include <render/export.h>

using namespace std;
using namespace std::chrono;

namespace pdp {

class View;
class NavigationPointAgent2d;
class ZoomPanAgent2d;
class ViewLayout;
class TimeSampleSequence;

/**
 * Rendering engine for 3D tomographic medical image volumes. This Class contains
 * all the objects such as OpenGL Context and volume container for rendering volumes.
 * The engine stores all the volumes and generates 2d images using an offscreen OpenGL Context
 *
 * @ingroup render
 */
class PDPRENDER_EXPORT VolumeRenderingEngine
{
public:
  enum LayoutOptions
  {
    LayoutOptionsFourViewChannels,
    LayoutOptionsFourViewOrientations,
    LayoutOptionsOneView
  };

  /**
   * Constructs a VolumeRenderingEngine
   */
  VolumeRenderingEngine(
    int width,
    int height,
    bool createJpegImage);

  /**
   * Destructs a VolumeTexture.
   */
  ~VolumeRenderingEngine();

  void setEndpoint(
    const std::string& remoteEndpointAddress,
    unsigned short remoteEndpointPort)
  {
    m_remoteEndpointAddress = remoteEndpointAddress;
    m_remoteEndpointPort = remoteEndpointPort;
  }

  /**
   * Returns the current remote endpoint address
   */
  const std::string& remoteEndpointAddress() const
  { return m_remoteEndpointAddress; }

  /**
   * Returns the current remote endpoint port
   */
  unsigned short remoteEndpointPort() const
  { return m_remoteEndpointPort; }

  /**
   * Returns the current engine information
   */
  std::string info() const
  { return m_info; }

  RenderContext* context()
  { return m_context; }

  /**
   * Returns the current context information
   */
  std::string contextInfo() const
  { return m_context->glInfo(); }

  LayoutOptions activeLayout() const
  { return m_activeLayout; }

  void setActiveLayout(
    LayoutOptions activeLayout)
  { m_activeLayout = activeLayout; }

  /**
   * Returns the current volume information
   */
  std::string volumeInfo() const
  { return m_volume->info(); }

  /**
   * Returns a reference to the current volume
   */
  VolumeR3* volume()
  { return m_volume; }

  /**
   * Loads a volume specified by the input full file path and name. The type of the volume is determined
   * by the file's extension. Nifiti ".nii" and Matlab ".mha" formats are currently supported
   *
   * @param filePathAndName
   * @return 'true' on successful volume load
   */
  int loadVolume(
    const std::string& filePathAndName);

  /**
   * Loads a volume group specified by the input full file path and name. Only Matlab (.mha) files are currently
   * supported in volume groups
   *
   * @param filePathAndName
   * @return 'true' on successful volume load
   */
  VolumeR3::DatasetType loadVolumeGroup(
    const std::string& filePathAndName);

  /**
   * Unloads all volumes and frees their data structures
   */
  void unloadVolumes();

  /**
   * Returns 'true' if rendering is enabled meaning that the context is valid
   */
  bool renderingEnabled() const
  { return m_context->renderingEnabled(); }

  /**
   * Returns 'true' if volumes are loaded and the volumeTexture has been successfully loaded
   */
  bool volumesLoaded() const
  { return m_volume->volumeLoaded(); }

  /**
   * Renders the current state of the rendering engine and updates the internal output image buffers with
   * the result.
   */
  void render(
    VolumeR3::VolumeChannel channel,
    VolumeR3::RenderOrientation renderOrientation,
    VolumeR3::StructureType structureType,
    float structureAlpha,
    float density,
    float brightness,
    bool useNavigationAgent);

  /**
   * Returns the current uncompressed image buffer.
   */
  const unsigned char* uncompressedImage()
  { return m_uncompressedImage; }

  int width() const
  { return m_context->width(); }

  int height() const
  { return m_context->height(); }

  const unsigned char* jpegImage() const
  { return m_jpegImage; }

  unsigned int jpegImageLength() const
  { return m_jpegImageLength; }

  void clearOverlay()
  { m_context->volumeTexture()->clearOverlay(); }

  void loadClassification()
  { m_context->volumeTexture()->loadClassificationFromVolume(); }

  void loadTruth()
  { m_context->volumeTexture()->loadTruthFromVolume(); }

  void updateContrast(
    int structureDisplayState)
  {
    if (structureDisplayState == 1)
      m_context->volumeTexture()->loadTruthFromVolume();
    else if (structureDisplayState == 2)
      m_context->volumeTexture()->loadClassificationFromVolume();
    else
      m_context->volumeTexture()->clearOverlay();
  }

  float renderTime() const
  { return (float)m_renderTime.count() * 1.0f; }

  float compressionTime() const
  { return (float)m_compressionTime.count() * 1.0f; }

  /**
   * Interprets mouse press events and passes them along to the views.
   * @a position is specified in window coordinates. @a doubleClick should
   * be true if this press event is the second one in a double click sequence
   * (mousePressEvent()should be be called also for the first one). Returns
   * @a true if the layout accepts the event, else it returns false.
   */
  bool mousePressEvent(
    const VectorR2i& position,
    MouseEvent::ButtonState button,
    MouseEvent::ModifierKeyState modifiers,
    MouseEvent::Interpretation interpretation,
    std::function<void()> updateRender);

  /**
   * Interprets mouse release events and passes them along to the views.
   * @a position is given in window coordinates. Returns @a true
   * if the layout accepts the event, else it returns false.
   */
  bool mouseReleaseEvent(
    const VectorR2i& position,
    MouseEvent::ButtonState button,
    MouseEvent::ModifierKeyState modifiers,
    std::function<void()> updateRender);

  /**
   * Interprets mouse motion events and passes them along to the views.
   * @a position is given in window coordinates. Returns @a true
   * if the layout accepts the event, else it returns false.
   */
  bool mouseMoveEvent(
    const VectorR2i& position,
    std::function<void()> updateRender);

private:
  std::string m_remoteEndpointAddress;
  unsigned short m_remoteEndpointPort;
  RenderContext* m_context;
  NavigationPointAgent2d* m_navigationAgent;
  View* m_views[5];
  ZoomPanAgent2d* m_zoomPanAgents[5];
  ViewLayout* m_viewLayouts[2];
  LayoutOptions m_activeLayout;
  VolumeR3* m_volume;
  std::string m_info;
  unsigned char* m_uncompressedImage;
  bool m_createJpegImage;
  unsigned char* m_jpegImage;
  unsigned int m_jpegImageLength;
#ifdef _WINDOWS
  chrono::duration<double, std::milli> m_renderTime;
  chrono::duration<double, std::milli> m_compressionTime;
  chrono::high_resolution_clock::time_point m_lastRenderTime;
#else
  std::chrono::duration<double, std::milli> m_renderTime;
  std::chrono::duration<double, std::milli> m_compressionTime;
  std::chrono::high_resolution_clock::time_point m_lastRenderTime;
#endif
  double m_renderingInterval;
  VolumeR3::DatasetType m_datasetType;
};

} // namespace pdp

