/** @file
 *
 * Stores a volume in OpenGL texture memory for 3D volume renderering
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <render/openglUtils.h>
#include <render/export.h>

namespace pdp {

class VolumeR3;
#ifdef PDP_NVIDIA_LOCAL
class VolumeCuda3d;
#endif

/**
 * Stores a volume in OpenGL texture memory for 3D volume rendering
 *
 * @ingroup render
 */
class PDPRENDER_EXPORT VolumeTexture3d
{
public:
  /**
   * Enumerated type used to identify the current state of the texture volume
   */
  enum VolumeTextureState
  {
    VolumeTextureEmpty = 0,
    VolumeTextureLoadedFixed,
    VolumeTextureLoadedMoving
  };

  /**
   * Constructs a VolumeTexture which stores a 3D volume from the input
   * volume in OpenGL texture memory. The VolumeTexture is initially empty
   */
  VolumeTexture3d();

  /**
   * Destructs a VolumeTexture.
   */
  ~VolumeTexture3d();

  /**
   * Returns the current state of the texture volume
   */
  VolumeTextureState state() const
  { return m_state; }

  /**
   * Returns the current requested state of the texture volume
   */
  VolumeTextureState requestedState() const
  { return m_requestedState; }

  /**
   * Loads a VolumeTexture from the input @a volume and the input @a lookupTable.
   * The resolution is based on the input @a volume size. This is not an immediate
   * load but simply saves the load parameters for the next display cycle.
   */
  void loadFromVolume(
    VolumeTextureState requestedState,
    VolumeR3* movingVolume,
    VolumeR3* fixedVolume,
    bool linearInterp);

  /**
   * Clears the structure overlay.
   */
  void clearOverlay();

  /**
   * Loads the classification stored in the current moving volume into the 3D texture
   * for display.
   */
  void loadClassificationFromVolume();

  /**
   * Loads the truth stored in the current moving volume into the 3D texture
   * for display.
   */
  void loadTruthFromVolume();

  /**
   * Returns a pointer to the current moving volume for the volume display texture. This
   * pointer can be nullptr and must be checked by the caller
   */
  VolumeR3* movingVolume()
  { return m_movingVolume; }

  /**
   * Unloads the texture and frees up the texture resources
   */
  void unload();

  /**
   * Returns @a true if the volume is currently loaded
   */
  bool volumeLoaded() const
  { return m_volumeLoaded; }

  /**
   * Returns the OpenGL pixel buffer object Id
   */
  GLuint glPbo() const
  { return m_glPbo; }

  /**
   * Returns the OpenGL 2d texture object Id
   */
  GLuint glAxialTexture2d() const
  { return m_glAxialTexture2d; }

  GLuint glSagittalTexture2d() const
  { return m_glSagittalTexture2d; }

  GLuint glCoronalTexture2d() const
  { return m_glCoronalTexture2d; }

  GLuint glMipTexture2d() const
  { return m_glMipTexture2d; }

  int mipSize() const
  { return m_mipSize; }

  void renderAxialSlice(
    VolumeR3::VolumeChannel channel,
    float positionRatio,
    VolumeR3::StructureType structureType,
    float structureAlpha);

  void renderSagittalSlice(
    VolumeR3::VolumeChannel channel,
    float positionRatio,
    VolumeR3::StructureType structureType,
    float structureAlpha);

  void renderCoronalSlice(
    VolumeR3::VolumeChannel channel,
    float positionRatio,
    VolumeR3::StructureType structureType,
    float structureAlpha);

  void renderMip(
    int channelType,
    int structureType,
    float structureAlpha,
    bool showStructures,
    const MatrixR4x4f& modelViewMatrix,
    float density,
    float brightness);

  /**
   * Immediate update of the texture. This method should be called at
   * render time to update the texture volume.
   */
  void updateTexture(
    bool clearOverlay,
    bool applyClassificationMask);

  /**
   * Immediate update of the texture parameters. This method should be called at
   * render time to update display properties as needed.
   */
  void updateProperties();

private:
  VolumeTextureState m_requestedState;
  VolumeTextureState m_state;
  VolumeR3* m_movingVolume;
  const VolumeR3* m_fixedVolume;
  bool m_volumeLoaded;
  GLuint m_glPbo;  // OpenGL pixel buffer object
  GLuint m_glAxialTexture2d; // OpenGL axial texture object
  GLuint m_glSagittalTexture2d; // OpenGL sagittal texture object
  GLuint m_glCoronalTexture2d; // OpenGL coronal texture object
  GLuint m_glMipTexture2d; // OpenGL MIP texture object
  int m_mipSize;
#ifdef PDP_NVIDIA_LOCAL
  VolumeCuda3d* m_volumeCuda3d;
#endif
  bool m_loadNeedsUpdate;
  bool m_propertiesNeedUpdate;
};

} // namespace pdp

