/** @file
 *
 * Zoom and pan agent for 2d views
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2020
 */

#pragma once

#include <render/Agent2d.h>

namespace pdp {

/**
 * Implements interactive, mouse controlled zooming and panning in a
 * View2d. It can only be attached to one View2d at a time.
 *
 * This controller updates the view-to-clip transformation
 * (that is, the @em projection) of the View2d it controls in order to adjust
 * and zoom. The scene-to-view transformation (that is, the camera position) is
 * adjusted in a direction orthogonal to the look direction to change the pan
 * position.
 *
 * A dragging operation with the left mouse button down will update
 * the pan so that the objects in the view seem to follow the
 * mouse. If a faster motion is desired, call #setPanSpeed with an
 * argument greater than one. For a slower motion, specify a speed
 * less than one.
 *
 * A dragging operation with the right mouse button
 * down will zoom the view by a factor that is proportional to the
 * vertical mouse motion. For faster zoom, call #setZoomSpeed with an
 * argument greater than one.
 *
 * @ingroup render
 */
class PDPRENDER_EXPORT ZoomPanAgent2d : virtual public Agent2d
{
public:
  /**
   * Constructs a ZoomPanAgent2d. Initially, this controller is not
   * attached to any view. It can be attached to at most one View2d with
   * View2D::insertRenderer.
   */
  ZoomPanAgent2d(
    float width = 1.0f,
    float height = 1.0f,
    float nearClipDistance = 0.2f,
    float farClipDistance = 10.0f);

  void addView(
    View* view)
  { m_views.push_back(view); }

  /**
   * Resets the zoom factor to one and the pan to zero.
   */
  void reset();

  /**
   * Sets the current window by assigning the width and height.
   */
  void setWindow(
    float width,
    float height);

  /**
   * Sets the current zoom factor. The view is zoomed towards the center
   * of the current viewing region. A @a zoom factor greater than 1 means
   * zoom in. A @a zoom less than 1 means zoom out.
   */
  void setZoom(
    float zoom);

  /**
   * Sets the zooming speed to @a zoomSpeed times faster than the default.
   */
  void setZoomSpeed(
    float zoomSpeed)
  { m_zoomSpeed = zoomSpeed; }

  /**
   * Returns the current zooming speed.
   */
  float zoomSpeed() const
  { return m_zoomSpeed; }

  bool mouseEvent(
    const MouseEvent& event,
    View& view,
    std::function<void()> updateRender) override;

private:
  double deltaZoom(
    const VectorR2f& from,
    const VectorR2f& to) const;

  void updateViewPosition(
    VectorR2f start,
    VectorR2f end);

  void updateViewRotation(
    VectorR2f start,
    VectorR2f end);

private:
  float m_width;
  float m_height;
  float m_nearClipDistance;
  float m_farClipDistance;
  float m_zoomSpeed;
  std::list<View*> m_views;

  // These are used while zooming or panning
  bool m_panning;
  bool m_zooming;
  VectorR2f m_downAt;
  float m_initialZoom;
};

} // namespace pdp

