/** @file
 * 
 * Gamepad classes and functions
 * 
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2019
 */
#pragma once

#include <thread>
#include <vector>
#include <string>
#include <plib/js.h>
#include <utilities/StoppableTask.h>
#include <utilities/export.h>

namespace pdp {

/**
 * Gamepad object
 *
 * @ingroup utilities
 */
class PDPUTILITIES_EXPORT Gamepad
{
public:
  /**
   * Constructs a Gamepad
   */
  explicit Gamepad(
    jsJoystick* js);

  /**
   * Destructs this Gamepad.
   */
  ~Gamepad();

  std::string name() const
  { return working() ? std::string(m_js->getName()) : std::string(); }

  int numberOfButtons() const
  { return m_js->getNumButtons(); }

  bool working() const
  { return (m_js != nullptr && !m_js->notWorking()) ? true : false; }

private:
  friend class GamepadsManagerPollingTask;
  friend class GamepadsManager;
  int m_buttonIdx;
  float* m_axes;
  jsJoystick* m_js;
};

class PDPUTILITIES_EXPORT GamepadsManagerPollingTask: public StoppableTask
{
public:
  GamepadsManagerPollingTask(
    unsigned int pollingIntervalInMilliSeconds,
    const std::vector<Gamepad*>& gamepads,
    std::function<void(int, int)>* buttonPressedFunction,
    std::function<void(int, float, float, float, float, float, float)>* axisChangedFunction,
    std::function<void()>* lostConnectFunction)
  : m_pollingIntervalInMilliSeconds(pollingIntervalInMilliSeconds),
    m_gamepads(gamepads),
    m_buttonPressedFunction(buttonPressedFunction),
    m_axisChangedFunction(axisChangedFunction),
    m_lostConnectFunction(lostConnectFunction)
  {}

  virtual ~GamepadsManagerPollingTask() {}

  void run();
  
private:
  unsigned int m_pollingIntervalInMilliSeconds;
  const std::vector<Gamepad*>& m_gamepads;
  std::function<void(int, int)>* m_buttonPressedFunction;
  std::function<void(int, float, float, float, float, float, float)>* m_axisChangedFunction;
  std::function<void()>* m_lostConnectFunction;
};

/**
 * Gamepads manager which detects, initializes, and manages attached gamepads
 *
 * @ingroup utilities
 */
class PDPUTILITIES_EXPORT GamepadsManager
{
public:
  static const int NumberOfAxises = 6;

  /**
   * Constructs a GamepadsManager
   */
  explicit GamepadsManager()
    : m_pollingTask(nullptr),
      m_pollingThread(nullptr)
  {}

  /**
   * Destructs this GamepadsManager.
   */
  ~GamepadsManager();

  /**
   * @returns the number of gamepads currently connteced
   */
  int numberOfGamepads() const
  { return (int)m_gamepads.size(); }

  /**
   * Releases all current gamepads and detects the current set of connected gamepads.
   *
   * @returns the number of gamepads found
   */
  int refreshGamepads();

  Gamepad* gamepad(
    int gamepadIdx)
  { return (gamepadIdx < (int)m_gamepads.size()) ? m_gamepads[gamepadIdx] : nullptr; }

  bool startPolling(
    unsigned int pollingIntervalInMilliSeconds,
    std::function<void(int, int)>* buttonPressedFunction,
    std::function<void(int, float, float, float, float, float, float)>* axisChangedFunction,
    std::function<void()>* lostConnectFunction);
  
  void stopPolling();
  
private:
  std::vector<Gamepad*> m_gamepads;
  GamepadsManagerPollingTask* m_pollingTask;
  std::thread* m_pollingThread;
};

} // namespace pdp
