/** @file
 *
 * String utility classes and functions
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <iosfwd>
#include <regex>
#include <cassert>
#include <stdexcept>
#include <cmath>
#include <stdio.h>
#include <string>
#include <iostream>
#include <sstream>
#include <vector>
#ifndef _WINDOWS
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wparentheses"
#endif
#include <boost/filesystem.hpp>
#ifndef _WINDOWS
#pragma GCC diagnostic pop
#endif
#include <utilities/export.h>

namespace pdp
{

/**
 * Union provides a mapping of different datatypes onto the same
 * raw memory space
 *
 * @ingroup utilities
 */
union basicTypesMap
{
  double double_val;
  float float_vals[2];
  unsigned int uints[2];
  int sints[2];
  short unsigned int ushorts[4];
  short int sshorts[4];
  unsigned char ubytes[8];
  char sbytes[8];
};

/**
 * Structure for MAC address information block
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT struct MACAddressInfoStruct
{
  std::string description;
  unsigned char address[6];
};

/**
 * Gets the local Media Access (MAC) address from all real network adaptors
 *
 * @ingroup utilities
 */
#ifndef PDP_OS_WINDOWS
PDPUTILITIES_EXPORT bool getLocalMACAddresses(
  std::vector<MACAddressInfoStruct>& macIds);
#endif

/**
 * Splits the input string into a vector of strings according to the input delimator charater
 *
 * @param s String to be split
 * @param delim Delimator character
 * @returns Vector of split strings
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT  std::vector<std::string> split(
  const std::string& s,
  char delim);

/**
 * Splits the input string into a vector of strings according to the input regular expression string
 *
 * @param s String to be split
 * @param rgxString regular expression string
 * @returns Vector of split strings
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT std::vector<std::string> split(
  const std::string& s,
  std::string rgxString);

/**
 * Removes duplicates from the input string list. This function alters the input string list in place
 *
 * @param stringList String list to remove duplicates from
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT void removeDuplicates(
  std::vector<std::string>& stringList);

/**
 * This function returns a string which is safe to insert into SQL commands and is equivelent
 * to the input string @a inString. The output string is created by doubling all single-quote and
 * backslash characters.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT std::string removedSqlCommandsString(
  const std::string& str);

/**
 * Trims the input string according to the input trim character (typically the 'white space' character)
 *
 * @param str String to be trimmed
 * @param trimCharacter Trim character. (Optional) Default value is whitespace
 * @returns The trimmed string
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT std::string trim(
  const std::string& str,
  char trimCharacter = 32);

/**
 * Trims the white space character from the input character string
 *
 * @param s String to be trimmed
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT void trimWhitespaceFromCharString(
  char *s);

/**
 * Parses the directory name or filename from the input full
 * path specification
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT bool getShortName(
  const std::string& fullPathname,
  std::string& shortName);

/**
 * Parses the extension from the input full
 * path specification
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT bool getExtension(
  const std::string& fullPathname,
  std::string& extension);

/**
 * Check for the presence of the test string in the input string
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT bool contains(
  const std::string& inputString,
  const std::string& testString);

/**
 * Returns a list of files in the input directory with the specified extension.
 * If the input extension is empty, all the files in the directory are returned.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT bool getFilesInDirectory(
  const std::string& fullPathname,
  const std::string& extension,
  std::vector<std::string>& filelist);

/**
 * Returns a list of directories in the input directory with the specified match string.
 * If the input match string is empty, all the directories in the directory are returned.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT bool getDirectoriesInDirectory(
  const std::string& fullPathname,
  const std::string& match,
  std::vector<std::string>& directorylist);

/**
 * Parses the full directory path from the input full path specification
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT bool getFullDirectoryPath(
  const std::string& fullPathname,
  std::string& fullDirectoryPath);

/**
 * Parses the parent directory name from the input full path specification
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT bool getParentDirectoryName(
  const std::string& fullPathname,
  std::string& parentDirectoryName);

/**
 * Returns whether the filesystem path <code>child</code> is a
 * descendant of the directory path <code>parent</code> (or equal to it).
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT bool isPathContainedBy(
  const std::string& child,
  const std::string& parent);
/**
 * Removes all instances of the target character @a targetChar from
 * the input string @a inputString.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT std::string removeCharacter(
  const std::string& inputString,
	const char targetChar);

/**
 * Removes all instances of the target sub-string @a targetString from
 * the input string @a inputString.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT std::string removeSubString(
  const std::string& inputString,
	const std::string& targetString);

/**
 * Adds the literal indicator '/' in front of all double quotes '"' in the string.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT std::string insertDoubleQuoteLiterals(
  const std::string& inputString);

/**
 * Writes the input line to the output stream followed by the newline character
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT void putLine(
  const std::string& line,
  std::ofstream& stream);

/**
 * Converts the input file from htlm to a C data structure.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT bool convertHtmlToStruct(
	const std::string& filename);

/**
 * Returns @a true if the input file @a filename can be opened for writing
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT bool writePermissions(
  const std::string& filename);

/**
 * Returns the length of the file stream passed in
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT unsigned long getFileStreamLength(
  std::ifstream& file);

/**
 * Parses the input text line @a line into a tag and value separated
 * by a semicolon
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT bool splitLine(
  const std::string& line,
  std::string& tag, std::string& value);

/**
 * Skips the next line in the input stream by reading and discarding the
 * next line.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT void skipLine(
  std::istream& inputStream);

/**
 * Skips the next specified number of lines @a numLines in the input
 * stream by reading and discarding each line.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT void skipLines(
  std::istream& inputStream,
  int numLines);

/**
 * Parses and returns the next value from the input stream. The value must
 * be preceded by a tag and a semi-colon in a text formatted stream.
 *
 * @ingroup utilities
 */
template <class T>
T getTaggedValue(
  std::istream& inputStream)
{
  std::string line;
  std::getline(inputStream, line);
  std::string tag;
  std::string valueString;
  splitLine(line, tag, valueString);
  std::istringstream valueStream(valueString);
  T value;
  if (!(valueStream >> value))
  {
    throw std::runtime_error("reading of value failed");
  }
  return value;
}

/**
 * Parses the input time string into number of seconds form the
 * beginning of the day.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT double getTimeInSeconds(
  const std::string& timeString);

/**
 * The method gets the next element of the specified type from the input stream and
 * decodes it from little endian format.
 *
 * @ingroup utilities
 */
template<typename E> E getR(
  std::istream& inputStream);

/**
 * The method gets the next bool from the input stream and
 * decodes it from little endian format.
 *
 * @ingroup utilities
 */
template<> PDPUTILITIES_EXPORT bool getR<bool>(
  std::istream& inputStream);

/**
 * The method gets the next char from the input stream and
 * decodes it from little endian format.
 *
 * @ingroup utilities
 */
template<> PDPUTILITIES_EXPORT char getR<char>(
  std::istream& inputStream);

/**
 * The method gets the next unsigned char from the input stream and
 * decodes it from little endian format.
 *
 * @ingroup utilities
 */
template<> PDPUTILITIES_EXPORT unsigned char getR<unsigned char>(
  std::istream& inputStream);

/**
 * The method gets the next short from the input stream and
 * decodes it from little endian format.
 *
 * @ingroup utilities
 */
template<> PDPUTILITIES_EXPORT short getR<short>(
  std::istream& inputStream);

/**
 * The method gets the next unsigned short from the input stream and
 * decodes it from little endian format.
 *
 * @ingroup utilities
 */
template<> PDPUTILITIES_EXPORT unsigned short getR<unsigned short>(
  std::istream& inputStream);

/**
 * The method gets the next int from the input stream and
 * decodes it from little endian format.
 *
 * @ingroup utilities
 */
template<> PDPUTILITIES_EXPORT int getR<int>(
  std::istream& inputStream);

/**
 * The method gets the next unsigned int from the input stream and
 * decodes it from little endian format.
 *
 * @ingroup utilities
 */
template<> PDPUTILITIES_EXPORT unsigned int getR<unsigned int>(
  std::istream& inputStream);

/**
 * The method gets the next float from the input stream and
 * decodes it from little endian format.
 *
 * @ingroup utilities
 */
template<> PDPUTILITIES_EXPORT float getR<float>(
  std::istream& inputStream);

/**
 * The method gets the next double from the input stream and
 * decodes it from little endian format.
 *
 * @ingroup utilities
 */
template<> PDPUTILITIES_EXPORT double getR<double>(
  std::istream& inputStream);

/**
 * The method gets the next std::string from the input stream and
 * decodes it from little endian format.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT void getStdString(
  std::istream& inputStream,
  std::string& s);

/**
 * The method gets the next encoded std::string from the input stream and
 * decodes it from little endian format. This encoding is for database version 10
 * and earlier.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT void getEncodedStdString(
  std::istream& inputStream,
  std::string& s);

/**
 * Gets the next encoded time_t from the input stream and decodes it in
 * little endian format.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT time_t getTime(
  std::istream& inputStream);

/**
 * The method encodes the input element of the specified type into
 * little endian format and writes it to the output stream.
 *
 * @ingroup utilities
 */
template<typename E>
void putR(
  E element,
  std::ostream& outputStream);

/**
 * The method encodes the input bool into
 * little endian format and writes it to the output stream.
 *
 * @ingroup utilities
 */
template<> PDPUTILITIES_EXPORT void putR<bool>(
  bool element,
  std::ostream& outputStream);

/**
 * The method encodes the input char into
 * little endian format and writes it to the output stream.
 *
 * @ingroup utilities
 */
template<> PDPUTILITIES_EXPORT void putR<char>(
  char element,
  std::ostream& outputStream);

/**
 * The method encodes the input unsigned char into
 * little endian format and writes it to the output stream.
 *
 * @ingroup utilities
 */
template<> PDPUTILITIES_EXPORT void putR<unsigned char>(
  unsigned char element,
  std::ostream& outputStream);

/**
 * The method encodes the input short into
 * little endian format and writes it to the output stream.
 *
 * @ingroup utilities
 */
template<> PDPUTILITIES_EXPORT void putR<short>(
  short element,
  std::ostream& outputStream);

/**
 * The method encodes the input unsigned short into
 * little endian format and writes it to the output stream.
 *
 * @ingroup utilities
 */
template<> PDPUTILITIES_EXPORT void putR<unsigned short>(
  unsigned short element,
  std::ostream& outputStream);

/**
 * The method encodes the input int into
 * little endian format and writes it to the output stream.
 *
 * @ingroup utilities
 */
template<> PDPUTILITIES_EXPORT void putR<int>(
  int element,
  std::ostream& outputStream);

/**
 * The method encodes the input unsigned int into
 * little endian format and writes it to the output stream.
 *
 * @ingroup utilities
 */
template<> PDPUTILITIES_EXPORT void putR<unsigned int>(
  unsigned int element,
  std::ostream& outputStream);

/**
 * The method encodes the input float into
 * little endian format and writes it to the output stream.
 *
 * @ingroup utilities
 */
template<> PDPUTILITIES_EXPORT void putR<float>(
  float element,
  std::ostream& outputStream);

/**
 * The method encodes the input double into
 * little endian format and writes it to the output stream.
 *
 * @ingroup utilities
 */
template<> PDPUTILITIES_EXPORT void putR<double>(
  double element,
  std::ostream& outputStream);

/**
 * The method encodes the input std::string into
 * little endian format and writes it to the output stream.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT void putStdString(
  const std::string& s,
  std::ostream& outputStream);

/**
 * The method encodes the input std::string into
 * encoded little endian format and writes it to the output stream.
 * This encoding is for database version 10 and earlier.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT void putEncodedStdString(
  const std::string& s,
  std::ostream& outputStream);

/**
 * Encodes the input time_t into a encoded little endian format and
 * writes it to the output stream.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT void putTime(
  const time_t& element,
  std::ostream& outputStream);

PDPUTILITIES_EXPORT int fileSize(
  const std::string& filename);

PDPUTILITIES_EXPORT void split(
  const std::string &s,
  char delim,
  std::vector<std::string> &elems);

/**
 * Compresses the input buffer and writes it to the output file using zlib.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT unsigned char* compressBufferToBuffer(
  unsigned int sourceBufferLength,
  int level,
  const unsigned char* sourceBuffer,
  unsigned long int& compressedBufferLength);

PDPUTILITIES_EXPORT bool compressBufferToFile(
  unsigned int sourceBufferLength,
  int level,
  const unsigned char* sourceBuffer,
  const std::string& fullFileName);

/**
 * Decompresses the input file and writes it to the output buffer using zlib.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT unsigned char* decompressBufferToBuffer(
  unsigned char* sourceBuffer,
  unsigned long int sourceBufferLength,
  unsigned long int uncompressedBufferLength);

PDPUTILITIES_EXPORT bool decompressFileToBuffer(
  const std::string& fullFileName,
  unsigned long int destBufferLength,
  unsigned char* destBuffer);

/**
 * Defines the error conditions that may be encountered when decompressing.
 *
 * @ingroup utilities
 */
enum ZlibUncompressResult
{
  InsufficientSpaceInDestination,  ///< Estimated size of decompressed data was wrong
  InsufficientSystemMemory,        ///< Allocation failed
  CompressedDataIsCorrupt,         ///< An error in the byte stream caused the decompression to fail crc
  SucessfulDecompression           ///< Successful decompression
};

/**
 * Functions for file access to support international characters. The windows implementation is
 * provided in a 3rd party lib so that dcmtk can be built using it.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT FILE* fopenUtf8(
  const char* filenameUtf8,
  const char* modeUtf8);

PDPUTILITIES_EXPORT int accessUtf8(
  const char* filenameUtf8,
  int accessMode);

PDPUTILITIES_EXPORT FILE* popenUtf8(
  const char* filenameUtf8,
  const char* modeUtf8);

PDPUTILITIES_EXPORT FILE* freopenUtf8(
  const char* filenameUtf8,
  const char* modeUtf8,
  FILE* file);

PDPUTILITIES_EXPORT void ifstreamOpenUtf8(
  const char* filenameUtf8,
  std::ifstream& streamObj,
  std::ios_base::openmode mode = std::ios_base::in);

PDPUTILITIES_EXPORT void ofstreamOpenUtf8(
  const char* filenameUtf8,
  std::ofstream& streamObj,
  std::ios_base::openmode mode = std::ios_base::out);

PDPUTILITIES_EXPORT std::string getenvUtf8(
  const char* varNameUtf8);

/**
 * The method gets the next element of the specified type from the input stream and
 * decodes it from little endian format.
 */
template<typename E>
E getR(
  std::istream& inputStream);

/**
 * The method gets the next bool from the input stream and
 * decodes it from little endian format.
 */
template<> PDPUTILITIES_EXPORT bool getR<bool>(
  std::istream& inputStream);

/**
 * The method gets the next char from the input stream and
 * decodes it from little endian format.
 */
template<> PDPUTILITIES_EXPORT char getR<char>(
  std::istream& inputStream);

/**
 * The method gets the next unsigned char from the input stream and
 * decodes it from little endian format.
 */
template<> PDPUTILITIES_EXPORT unsigned char getR<unsigned char>(
  std::istream& inputStream);

/**
 * The method gets the next short from the input stream and
 * decodes it from little endian format.
 */
template<> PDPUTILITIES_EXPORT short getR<short>(
  std::istream& inputStream);

/**
 * The method gets the next unsigned short from the input stream and
 * decodes it from little endian format.
 */
template<> PDPUTILITIES_EXPORT unsigned short getR<unsigned short>(
  std::istream& inputStream);

/**
 * The method gets the next int from the input stream and
 * decodes it from little endian format.
 */
template<> PDPUTILITIES_EXPORT int getR<int>(
  std::istream& inputStream);

/**
 * The method gets the next unsigned int from the input stream and
 * decodes it from little endian format.
 */
template<> PDPUTILITIES_EXPORT unsigned int getR<unsigned int>(
  std::istream& inputStream);

/**
 * The method gets the next float from the input stream and
 * decodes it from little endian format.
 */
template<> PDPUTILITIES_EXPORT float getR<float>(
  std::istream& inputStream);

/**
 * The method gets the next double from the input stream and
 * decodes it from little endian format.
 */
template<> PDPUTILITIES_EXPORT double getR<double>(
  std::istream& inputStream);

/**
 * Returns the windows username, or the userid on MacOSX and Linux.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT std::string getUserName();

/**
 * Returns the home directory path.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT std::string getUserHome();

/**
 * Returns a string representation of the input number up to the specified
 * significant digits.  For very small numbers where the significant digits
 * are not enough to capture the actual value of the number, the scientific
 * notation is used in order to avoid returning a string value of "0.0".
 *
 * @ingroup utilities
*/
PDPUTILITIES_EXPORT std::string formatNumber(
  double number,
  int significantDigits,
  double notationConversionValue = 0.001,
  bool forcePlusSign = false);

PDPUTILITIES_EXPORT std::string formatNumber(
  float number,
  int significantDigits,
  float notationConversionValue = 0.001f,
  bool forcePlusSign = false);

/**
 * Computes max precision possible for formatting a floating point value
 * that keeps the resulting string within 16 characters/bytes as required
 * by DICOM's Decimal String (DS) value representation.
 *
 * @ingroup utilities
 */
template <typename T>
int formattingPrecisionNeeded(T val)
{
  // compute base-10 exponent as it would be shown in scientific notation
  int exp = (std::fpclassify(val) == FP_ZERO) ? 0 : static_cast<int>(std::log10(std::fabs(val)));
  double fraction = val * std::pow(10, -exp);
  if (std::fabs(fraction) < 1.0)
    exp -= 1;

  // max precision assuming dot '.' unless all integer digits will be used
  int precision;
  if (val >= 0.0)
  {
    precision = val >= 1e15 && val < 1e16 ? 16 : 15;
  }
  else
  {
    // one less than positive for '-'
    precision = val > -1e15 && val <= -1e14 ? 15 : 14;
  }

  // reduce as needed if scientific notation will be triggered,
  // default formatting is %g and fprintf docs state:
  //   Let P equal the precision if nonzero, 6 if the precision is not specified, or 1 is the precision is 0.
  //   Then, if a conversion with style E would have an exponent of X :
  //      if P > X > -4, the conversion is with style f or F and precision P - 1 - X.
  //      otherwise, the conversion is with style e or E and precision P - 1.
  if (exp < -99)
    precision -= 5; // "e-123"
  else if (exp < -4)
    precision -= 4; // "e-01"
  else if (exp >= precision)
  {
    if (exp > 99)
      precision -= 5; // "e+123"
    else
      precision -= 4; // "e+12"
  }
  return std::min(std::numeric_limits<T>::max_digits10, precision);
}

/**
 * Formats a floating point number to text with "full" precision.
 * Note this is intended for DICOM and therefor is slightly lossy
 * with doubles or larger as DICOM has a 16-byte max length and we
 * need to make room for the '.', '-', 'e', and exponent, whereas
 * we need precision of 17 to safely round-trip a double.
 *
 * @ingroup utilities
 */
template <typename T>
std::string formatFullFloatingPoint(T val)
{
  std::ostringstream oss;
  oss.precision(formattingPrecisionNeeded(val));
  oss << val;
  std::string result = oss.str();
  return result;
}

/**
 * Formats a series of floating point numbers to text with "full" precision.
 * Note this is intended for DICOM and therefor is slightly lossy
 * with doubles or larger as DICOM has a 16-byte max length and we
 * need to make room for the '.', '-', 'e', and exponent whereas we need precision
 * of 17 to safely round-trip a double.
 *
 * Values are joined with the given separator, defaulting to back-slash
 * as used in DICOM.
 *
 * @ingroup utilities
 */
template <typename T>
std::string formatFullFloatingPoints(
  std::initializer_list<T> vals,
  const std::string& sep = "\\")
{
  std::ostringstream oss;
  bool notfirst = false;
  for (auto val : vals)
  {
    if (notfirst)
      oss << sep;
    oss.precision(formattingPrecisionNeeded(val));
    oss << val;
    notfirst = true;
  }
  return oss.str();
}

/**
 * The following function will encode a string to Base64.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT std::string toBase64(
  const std::string& text);

PDPUTILITIES_EXPORT std::string intToPaddedString(
  int value);

/**
 * Creates a group file and rearranges the data files for a BRATS2015 organized patient with
 * matlab (.mha) files.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT bool createBrats2015MhaVolumeGroupFile(
  const std::string& patientDirectoryFullPath);

/**
 * Creates a group file and rearranges the data files for a BRATS2015 organized patient with
 * matlab (.mha) files.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT bool createBrats2015MhaVolumeGroupFile(
  const std::string& patientDirectoryFullPath);

} // namespace pdp
