/** @file
 *
 * Time utility classes and functions
 *
 * @author Paul Pantalone
 * @copyright Paul Pantalone. All Rights Reserved
 * @date 2018
 */
#pragma once

#include <stdio.h>
#include <string>
#include <chrono>
#ifdef PDP_OS_WINDOWS
#include <time.h>
#else
#include <sys/time.h>
#endif
#include <utilities/export.h>

namespace pdp
{
/**
 * Inputs a string in the UTC time format of 'yyyy-MM-ddTHH:mm:ss.mmmZ' and returns a time object
 * set to the time value of the input string
 *
 * @param timeStr Time string with the following time format: 'yyyy-MM-ddTHH:mm:ss.mmmZ'
 * @returns Time object set to the time value of the input string
 *
 * @note This function requires that the input string have the form 'yyyy-MM-ddTHH:mm:ss.mmmZ'.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT std::chrono::system_clock::time_point utcTimeStringToTimePoint(
  const std::string& timeStr);

/**
 * Inputs a string in the format of 'yyyy.mm.dd-HH.MM.SS.fff' and returns a time object
 * set to the time value of the input string
 *
 * @param timeStr Time string with the following time format: 'yyyy.mm.dd-HH.MM.SS.fff'
 * @returns Time object set to the time value of the input string
 *
 * @note This function requires that the input string have the form 'yyyy.mm.dd-HH.MM.SS.fff'.
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT std::chrono::system_clock::time_point stringToTimePoint(
  const std::string& timeStr);

/**
 * Inputs a time object and returns a string representing
 * the value of the input time object with the following format: 'yyyy.mm.dd-HH.MM.SS.fff'
 *
 * @param tp Time object
 * @returns String with the following time format: 'yyyy.mm.dd-HH.MM.SS.fff'
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT std::string timePointToString(
  const std::chrono::system_clock::time_point& tp);

/**
 * Inputs a two time objects and returns a float with the difference value in seconds
 *
 * @param tp2 Time object
 * @param tp1 Time object
 * @returns float with the difference between the time point in seconds
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT float timeDifferenceInSeconds(
  const std::chrono::system_clock::time_point& tp2,
  const std::chrono::system_clock::time_point& tp1);

/**
 * Inputs a two time objects and returns a string representing
 * the difference value between them with the following format: 'dddd-HH:MM:SS:mmm.mmm'
 *
 * @param tp2 Time object
 * @param tp1 Time object
 * @returns String with the following time format: 'dddd-HH:MM:SS:mmm.mmm'
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT std::string timePointDifference(
  const std::chrono::system_clock::time_point& tp2,
  const std::chrono::system_clock::time_point& tp1);

/**
 * Inputs a two time objects and returns a unsigned long int with
 * the difference value between them in micro-seconds
 *
 * @param tp2 Time object
 * @param tp1 Time object
 * @returns unsigned long int time difference in micro-seconds
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT unsigned long int timePointDifferenceInMicroSeconds(
  const std::chrono::high_resolution_clock::time_point& tp2,
  const std::chrono::high_resolution_clock::time_point& tp1);

/**
 * Inputs a two time objects and returns a string representing
 * the difference value between them with the following format: 'HH:MM:SS:mmm.mmm'
 *
 * @param tp2 Time object
 * @param tp1 Time object
 * @returns String with the following time format: 'HH:MM:SS:mmm.mmm'
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT std::string timePointDifferenceHighResolution(
  const std::chrono::high_resolution_clock::time_point& tp2,
  const std::chrono::high_resolution_clock::time_point& tp1);

/**
 * Inputs number of micro seconds and returns a string representing
 * the value in the following time format: 'HH:MM:SS:mmm.mmm'
 *
 * @param tp2 Time object
 * @param tp1 Time object
 * @returns String with the following time format: 'HH:MM:SS:mmm.mmm'
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT std::string microSecondsToTimeString(
  unsigned long int numMicroSeconds,
  bool markZero = false);

PDPUTILITIES_EXPORT std::string secondsToTimeString(
  double numSeconds);

/**
 * Converts the input time in clock_t to seconds
 *
 * @ingroup utilities
 */
PDPUTILITIES_EXPORT double clockToSeconds(
  clock_t clockTime);

} // namespace pdp
